import logging
import time
from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum, IntEnum, auto
from typing import TYPE_CHECKING

from pandas import DataFrame, Series
from tqdm import tqdm

from mstrio import config
from mstrio.api import datasources as datasources_api
from mstrio.api import monitors, objects, projects
from mstrio.connection import Connection
from mstrio.helpers import IServerError, VersionException
from mstrio.server.project_languages import (
    DataLanguageSettings,
    MetadataLanguageSettings,
)
from mstrio.utils import helper, time_helper
from mstrio.utils.entity import DeleteMixin, Entity, EntityBase, ObjectTypes
from mstrio.utils.enum_helper import AutoName, AutoUpperName, get_enum_val
from mstrio.utils.resolvers import validate_owner_key_in_filters
from mstrio.utils.response_processors import datasources as datasources_processors
from mstrio.utils.response_processors import objects as objects_processors
from mstrio.utils.response_processors import projects as projects_processors
from mstrio.utils.settings.base_settings import BaseSettings
from mstrio.utils.time_helper import DatetimeFormats
from mstrio.utils.version_helper import (
    is_server_min_version,
    meets_minimal_version,
    method_version_handler,
)
from mstrio.utils.vldb_mixin import ModelVldbMixin
from mstrio.utils.wip import wip

if TYPE_CHECKING:
    from requests import Response

    from mstrio.datasources import DatasourceInstance
    from mstrio.server.environment import Environment
    from mstrio.server.node import Node
    from mstrio.users_and_groups import User

logger = logging.getLogger(__name__)

SYSTEM_OBJECTS_VERSION_PROPERTY_SET_ID = "C02146A211D46D60C0001786395B684F"
AE_VERSION_PROPERTY_INDEX = 8


class ProjectStatus(IntEnum):
    ACTIVE = 0
    ERRORSTATE = -3
    EXECIDLE = 1
    FULLIDLE = 7
    OFFLINE = -1
    OFFLINEPENDING = -2
    ONLINEPENDING = -4
    REQUESTIDLE = 2
    WHEXECIDLE = 4


class IdleMode(AutoName):
    """Used to specify the exact behaviour of `idle` request.

    `REQUEST` (Request Idle): all executing and queued jobs finish executing
        and any newly submitted jobs are rejected.
    `EXECUTION` (Execution Idle for All Jobs): all executing, queued, and
        newly submitted jobs are placed in the queue, to be executed when
        the project resumes.
    `WAREHOUSEEXEC` (Execution Idle for Warehouse jobs): all executing,
        queued, and newly submitted jobs that require SQL to be submitted
        to the data warehouse are placed in the queue, to be executed when
        the project resumes. Any jobs that do not require SQL to be
        executed against the data warehouse are executed.
    `FULL` (Request Idle and Execution Idle for All jobs): all executing and
        queued jobs are canceled, and any newly submitted jobs are rejected.
    `PARTIAL` (Request Idle and Execution Idle for Warehouse jobs): all
        executing and queued jobs that do not submit SQL against the data
        warehouse are canceled, and any newly submitted jobs are rejected.
        Any currently executing and queued jobs that do not require SQL to
        be executed against the data warehouse are executed.
    """

    REQUEST = "request_idle"
    EXECUTION = "exec_idle"
    WAREHOUSEEXEC = "wh_exec_idle"
    FULL = "partial_idle"
    PARTIAL = "full_idle"
    UNLOADED = auto()
    LOADED = auto()
    UNLOADED_PENDING = auto()
    LOADED_PENDING = auto()
    PENDING = auto()
    UNKNOWN = auto()


class LockType(AutoName):
    """Enum representing the type of lock applied to a project.

    `TEMPORAL_INDIVIDUAL`: A temporary lock that restricts all other sessions
        except the current user's session from editing the project.
        This lock disappears when the user's session expires.
    `TEMPORAL_CONSTITUENT`: A temporary lock that restricts all other sessions
        except the current user's session from editing the project and all
        objects in the project. This lock disappears when the user's session
        expires.
    `PERMANENT_INDIVIDUAL`: A permanent lock that prevents all users from
        editing the project. This lock does not expire and must be removed
        before the project can be edited again.
    `PERMANENT_CONSTITUENT`: A permanent lock that prevents all users from
        editing the project and all objects in the project. This lock does not
        expire and must be removed before the project and its objects can be
        edited again.
    `NOT_LOCKED`: Represents the state where the project is not locked
        and can be edited by users.
    """

    TEMPORAL_INDIVIDUAL = auto()
    TEMPORAL_CONSTITUENT = auto()
    PERMANENT_INDIVIDUAL = auto()
    PERMANENT_CONSTITUENT = auto()
    NOT_LOCKED = auto()


def compare_project_settings(
    projects: list["Project"], show_diff_only: bool = False
) -> DataFrame:
    """Compares settings of project objects.

    Args:
        projects (List[Project]): List of project objects
            to compare.
        show_diff_only (bool, optional): Whether to display all settings
            or only different from first project in list.
    """
    to_be_compared = {}
    df = DataFrame({}, columns=['initial'])

    for project in projects:
        to_be_compared[project.name] = project.settings.to_dataframe().reset_index()
        if df.empty:
            df.columns = ['setting']
            df['setting'] = to_be_compared[project.name]['setting']
        df[project.name] = to_be_compared[project.name]['value']

    if show_diff_only:
        index = Series([True] * len(df['setting']))
        base = projects[0].name
        for project_name in to_be_compared.keys():
            compare = df[base].eq(df[project_name])
            index = compare & index
        df = df[~index]
        if df.empty and config.verbose:
            project_names = list(to_be_compared.keys())
            project_names.remove(base)
            msg = (
                f"There is no difference in settings between project '{base}' and "
                f"remaining projects: '{project_names}'"
            )
            logger.info(msg)
    return df


def list_projects(
    # allows for `conn` to be Env to support `list_projects(env)` syntax
    conn: "Connection | Environment | None" = None,
    env: "Environment | None" = None,
    *args,
    **kwargs,
):
    """Return list of project objects or project dicts if `to_dictionary=True`.
    Optionally filter the Projects by specifying the filters keyword arguments.

    This is a helper method using `Environment.list_projects()` under the hood.
    See `Environment.list_projects()` documentation for more details.

    Either `conn` or `env` must be provided.

    Attributes:
        conn (Connection, optional): A Strategy One connection object.
        env (Environment, optional): Environment object instance.
        ... (optional): Additional parameters to pass to
            `Environment.list_projects()`.
    """
    if conn is None and env is None:
        raise ValueError("Either connection or environment must be provided.")

    from mstrio.server.environment import Environment

    if not env and isinstance(conn, Environment):
        env = conn

    if env:
        return env.list_projects(*args, **kwargs)

    return Project._list_projects(conn, *args, **kwargs)


class Project(Entity, ModelVldbMixin, DeleteMixin):
    """Object representation of Strategy One Project (Project) object.

    Attributes:
        connection: A Strategy One connection object
        settings: Project settings object
        id: Project ID
        name: Project name
        description: Project description
        alias: Project alias
        type: Object type
        subtype: Object subtype
        ext_type: Object extended type
        nodes: List of nodes on which project is loaded
        date_created: Creation time, "yyyy-MM-dd HH:mm:ss" in UTC
        date_modified: Last modification time, "yyyy-MM-dd
        version: Version ID
        owner: owner ID and name
        acg: Access rights (See EnumDSSXMLAccessRightFlags for possible values)
        acl: Object access control list
        status: Project
        ancestors: List of ancestor folders
        data_language_settings: Project data language settings
        metadata_language_settings: Project metadata language settings
    """

    @dataclass
    class LockStatus(helper.Dictable):
        """Object representation of Project Lock Status.

        Attributes:
            lock_type: Lock type
            lock_time: Lock time
            comment: Lock comment
            machine_name: Machine name
            owner: User object
        """

        @staticmethod
        def _parse_owner(source, connection):
            """Parses owner from the API response."""
            from mstrio.users_and_groups import User

            return User.from_dict(source, connection)

        _FROM_DICT_MAP = {
            'lock_type': LockType,
            'lock_time': DatetimeFormats.FULLDATETIME,
            'owner': _parse_owner,
        }

        lock_type: LockType
        lock_time: datetime | None = None
        comment: str | None = None
        machine_name: str | None = None
        owner: 'User | None' = None

    _OBJECT_TYPE = ObjectTypes.PROJECT
    _API_GETTERS = {
        **Entity._API_GETTERS,
        ('status', 'alias'): projects.get_project,
        'nodes': monitors.get_node_info,
        (
            'data_language_settings',
            'metadata_language_settings',
        ): projects_processors.get_project_internalization,
        'lock_status': projects_processors.get_project_lock,
    }
    _API_PATCH = {
        **Entity._API_PATCH,
        'alias': (projects.update_project, 'patch'),
        (
            'comments',
            'owner',
        ): (objects_processors.update, 'partial_put'),
    }
    _PATCH_PATH_TYPES = {
        **Entity._PATCH_PATH_TYPES,
        'alias': str,
    }
    _API_DELETE = staticmethod(projects.delete_project)
    _FROM_DICT_MAP = {
        **Entity._FROM_DICT_MAP,
        'status': ProjectStatus,
        'data_language_settings': DataLanguageSettings.from_dict,
        'metadata_language_settings': MetadataLanguageSettings.from_dict,
        'lock_status': LockStatus.from_dict,
    }
    _STATUS_PATH = "/status"
    _MODEL_VLDB_API = {
        'GET_ADVANCED': projects.get_vldb_settings,
        'PUT_ADVANCED': projects.update_vldb_settings,
        'GET_APPLICABLE': projects.get_applicable_vldb_settings,
    }
    _REST_ATTR_MAP = {
        'data': 'data_language_settings',
        'metadata': 'metadata_language_settings',
    }

    def __init__(
        self,
        connection: Connection,
        name: str | None = None,
        id: str | None = None,
    ) -> None:
        """Initialize Project object by passing `name` or `id`. When `id` is
        provided (not `None`), `name` is omitted.

        Args:
            connection: Strategy One connection object returned
                by `connection.Connection()`
            name: Project name
            id: Project ID
        """

        # initialize either by ID or Project Name
        if id is None and name is None:
            helper.exception_handler(
                "Please specify either 'name' or 'id' parameter in the constructor.",
                exception_type=ValueError,
            )

        if id is None:
            project_list = Project._list_project_ids(connection, name=name)
            if project_list:
                id = project_list[0]
            else:
                helper.exception_handler(
                    f"There is no project with the given name: '{name}'",
                    exception_type=ValueError,
                )

        try:
            super().__init__(connection=connection, object_id=id, name=name)
        except IServerError as e:
            if not self.is_loaded():
                helper.exception_handler(
                    (
                        "Some projects are either unloaded or idled. Change "
                        "status using the 'load()' or 'resume()' method to use "
                        "all functionality."
                    ),
                    exception_type=UserWarning,
                )
            else:
                raise e

    def _init_variables(self, **kwargs) -> None:
        super()._init_variables(**kwargs)
        self._status = kwargs.get("status")
        self._alias = kwargs.get("alias")
        self._nodes = kwargs.get("nodes")
        self._data_language_settings = kwargs.get("data_language_settings")
        self._metadata_language_settings = kwargs.get("metadata_language_settings")
        self._lock_status = kwargs.get("lock_status")

    @classmethod
    def _create(
        cls,
        connection: Connection,
        name: str,
        description: str | None = None,
        force: bool = False,
    ) -> 'Project | None':
        user_input = 'N'
        if not force:
            user_input = input(
                f"Are you sure you want to create new project '{name}'? [Y/N]: "
            )

        if force or user_input == 'Y':
            # Create new project
            with tqdm(
                desc=f"Please wait while Project '{name}' is being created.",
                bar_format='{desc}',
                leave=False,
                disable=config.verbose,
            ):
                projects.create_project(connection, name, description)
                http_status, i_server_status = 500, 'ERR001'
                while http_status == 500 and i_server_status == 'ERR001':
                    time.sleep(1)
                    response = projects.get_project(
                        connection, name, whitelist=[('ERR001', 500)]
                    )
                    http_status = response.status_code
                    data = response.json()
                    i_server_status = data.get('code')
            if config.verbose:
                logger.info(f"Project '{name}' successfully created.")
            return cls.from_dict(data, connection=connection)
        else:
            return None

    @classmethod
    @method_version_handler('11.2.0000')
    def _list_projects(
        cls,
        connection: Connection,
        to_dictionary: bool = False,
        limit: int | None = None,
        **filters,
    ) -> list["Project"] | list[dict]:
        validate_owner_key_in_filters(filters)

        msg = "Error getting information for a set of Projects."
        objects = helper.fetch_objects_async(
            connection,
            monitors.get_projects,
            monitors.get_projects_async,
            dict_unpack_value='projects',
            limit=limit,
            chunk_size=1000,
            error_msg=msg,
            filters=filters,
        )
        if to_dictionary:
            return objects
        else:
            projects = [
                cls.from_dict(source=obj, connection=connection) for obj in objects
            ]
            projects_loaded = Project._list_loaded_projects(
                connection, to_dictionary=True
            )
            projects_loaded_ids = [project['id'] for project in projects_loaded]
            unloaded = [
                project for project in projects if project.id not in projects_loaded_ids
            ]

            if unloaded:
                msg = (
                    f"Projects {[project.name for project in unloaded]} are either "
                    f"unloaded or idled. Change status using the 'load()' or 'resume()'"
                    f" method to use all functionality."
                )
                helper.exception_handler(msg, exception_type=UserWarning)
            return projects

    @classmethod
    def _list_project_ids(
        cls, connection: Connection, limit: int | None = None, **filters
    ) -> list[str]:
        project_dicts = Project._list_projects(
            connection=connection,
            to_dictionary=True,
            limit=limit,
            **dict(filters),
        )
        return [project['id'] for project in project_dicts]

    @classmethod
    def _list_loaded_projects(
        cls, connection: Connection, to_dictionary: bool = False, **filters
    ) -> list["Project"] | list[dict]:
        validate_owner_key_in_filters(filters)

        response = projects.get_projects(connection, whitelist=[('ERR014', 403)])
        list_of_dicts = response.json() if response.ok else []
        list_of_dicts = helper.camel_to_snake(list_of_dicts)  # Convert keys
        raw_project = helper.filter_list_of_dicts(list_of_dicts, **filters)

        if to_dictionary:
            # return list of project names
            return raw_project
        else:
            # return list of Project objects
            return [
                cls.from_dict(source=obj, connection=connection) for obj in raw_project
            ]

    def alter(
        self,
        name: str | None = None,
        description: str | None = None,
        comments: str | None = None,
        owner: 'str | User | None' = None,
        alias: str | None = None,
    ) -> None:
        """Alter project name or/and description.

        Args:
            name (str, optional): new name of the project.
            description (str, optional): new description of the project.
            comments (str, optional): long description of the project.
            owner: (str, User, optional): owner of the project.
            alias (str, optional): new alias of the project.
        """
        from mstrio.users_and_groups import User

        if alias and not is_server_min_version(self.connection, '11.5.0300'):
            msg = (
                "Environments must run IServer version `11.5.0300` or newer. "
                "Please update your environments to alter project alias."
            )
            raise VersionException(msg)
        if isinstance(owner, User):
            owner = owner.id

        properties = helper.filter_params_for_func(
            self.alter, locals(), exclude=['self']
        )

        self._alter_properties(**properties)

    def __change_project_state(
        self, func, on_nodes: str | list[str] | None = None, **mode
    ):
        if isinstance(on_nodes, list):
            for node in on_nodes:
                func(node, **mode)
        elif isinstance(on_nodes, str):
            func(on_nodes, **mode)
        elif on_nodes is None:
            for node in self.nodes:
                func(node.get('name'), **mode)  # type: ignore
        else:
            helper.exception_handler(
                "'on_nodes' argument needs to be of type: [list[str], str, NoneType]",
                exception_type=TypeError,
            )

    def _change_project_state_all_nodes(
        self, status: str, delete_sessions: bool | None = None
    ):
        body = {"status": status}
        response = monitors.update_project_status(
            self.connection,
            body,
            project_id=self.id,
            delete_sessions=delete_sessions,
        )
        if response.status_code == 202 and config.verbose:
            logger.info(
                f"Project '{self.id}' changed status to '{status}' on all nodes."
            )

    @method_version_handler('11.2.0000')
    def idle(
        self,
        on_nodes: str | list[str] | None = None,
        mode: IdleMode | str = IdleMode.REQUEST,
    ) -> None:
        """Request to idle a specific cluster node. Idle project with mode
        options.

        Args:
            on_nodes: Name of node, if not passed, project will be idled on
                all of the nodes.
            mode: One of: `IdleMode` values.
        """

        def idle_project(node: str, mode: IdleMode):
            body = {
                "operationList": [
                    {"op": "replace", "path": self._STATUS_PATH, "value": mode.value}
                ]
            }
            response = monitors.update_node_properties(
                self.connection, node, self.id, body
            )
            if response.status_code == 202:
                tmp = helper.filter_list_of_dicts(self.nodes, name=node)
                tmp[0]['projects'] = [response.json()['project']]
                self._nodes = tmp
                if tmp[0]['projects'][0]['status'] != mode.value:
                    self.fetch('nodes')
                if config.verbose:
                    logger.info(
                        f"Project '{self.id}' changed status to '{mode}' on node "
                        f"'{node}'."
                    )

        if not isinstance(mode, IdleMode):
            # Previously `mode` was just a string with possible values
            # corresponding to the member names of the current IdleMode enum.
            # This attempts to convert it to avoid breaking backwards compat.
            if mode in IdleMode.__members__.values():
                mode = IdleMode(mode)
            elif mode in IdleMode.__members__:
                mode = IdleMode[mode]
            else:
                helper.exception_handler(
                    "Unsupported mode, please provide a valid `IdleMode` value.",
                    KeyError,
                )

        self.__change_project_state(func=idle_project, on_nodes=on_nodes, mode=mode)

    @method_version_handler('11.2.0000')
    def resume(self, on_nodes: str | list[str] | None = None) -> None:
        """Request to resume the project on the chosen cluster nodes. If
        nodes are not specified, the project will be loaded on all nodes.

        Args:
            on_nodes: Name of node, if not passed, project will be resumed
                on all of the nodes.
        """

        def resume_project(node):
            body = {
                "operationList": [
                    {"op": "replace", "path": self._STATUS_PATH, "value": "loaded"}
                ]
            }
            response = monitors.update_node_properties(
                self.connection, node, self.id, body
            )
            if response.status_code == 202:
                tmp = helper.filter_list_of_dicts(self.nodes, name=node)
                tmp[0]['projects'] = [response.json()['project']]
                self._nodes = tmp
                if tmp[0]['projects'][0]['status'] != 'loaded':
                    self.fetch('nodes')
                if config.verbose:
                    logger.info(f"Project '{self.id}' resumed on node '{node}'.")

        self.__change_project_state(func=resume_project, on_nodes=on_nodes)

    @staticmethod
    def _normalize_nodes(
        on_nodes: 'str | list[str] | Node | list[Node] | None',
    ) -> str | list[str] | None:
        from mstrio.server.node import Node

        if isinstance(on_nodes, list):
            return [node.name if isinstance(node, Node) else node for node in on_nodes]
        return on_nodes.name if isinstance(on_nodes, Node) else on_nodes

    def _load_project(self, node):
        body = {
            "operationList": [
                {"op": "replace", "path": self._STATUS_PATH, "value": "loaded"}
            ]
        }
        response = monitors.update_node_properties(self.connection, node, self.id, body)
        if response.status_code == 202:
            tmp = helper.filter_list_of_dicts(self.nodes, name=node)
            tmp[0]['projects'] = [response.json()['project']]
            self._nodes = tmp
            if tmp[0]['projects'][0]['status'] != 'loaded':
                self.fetch('nodes')
            if config.verbose:
                logger.info(f"Project '{self.id}' loaded on node '{node}'.")

    @method_version_handler('11.2.0000')
    def load(
        self, on_nodes: 'str | list[str] | Node | list[Node] | None' = None
    ) -> None:
        """Request to load the project onto the chosen cluster nodes. If
        nodes are not specified, the project will be loaded on all nodes.

        Args:
            on_nodes: Name of node, if not passed, project will be loaded
                on all of the nodes.
        """
        on_nodes = self._normalize_nodes(on_nodes)
        all_nodes_are_selected = on_nodes is None or {
            node['name'] for node in self.nodes
        } == set(on_nodes)

        if all_nodes_are_selected and meets_minimal_version(
            self.connection.iserver_version, '11.5.0500'
        ):
            self._change_project_state_all_nodes(status="loaded")
        else:
            self.__change_project_state(func=self._load_project, on_nodes=on_nodes)

    def _unload_project(self, node):
        body = {
            "operationList": [
                {"op": "replace", "path": self._STATUS_PATH, "value": "unloaded"}
            ]
        }
        response = monitors.update_node_properties(
            self.connection, node, self.id, body, whitelist=[('ERR001', 500)]
        )
        if response.status_code == 202:
            tmp = helper.filter_list_of_dicts(self.nodes, name=node)
            tmp[0]['projects'] = [response.json()['project']]
            self._nodes = tmp
            if tmp[0]['projects'][0]['status'] != 'unloaded':
                self.fetch('nodes')
            if config.verbose:
                logger.info(f"Project '{self.id}' unloaded on node '{node}'.")
        if response.status_code == 500 and config.verbose:  # handle whitelisted
            logger.warning(f"Project '{self.id}' already unloaded on node '{node}'.")

    @method_version_handler('11.2.0000')
    def unload(
        self,
        on_nodes: 'str | list[str] | Node | list[Node] | None' = None,
        delete_sessions: bool | None = None,
    ) -> None:
        """Request to unload the project from the chosen cluster nodes. If
        nodes are not specified, the project will be unloaded on all nodes.
        The unload action cannot be performed until all jobs and connections
        for project are completed. Once these processes have finished,
        pending project will be automatically unloaded.

        Args:
            on_nodes: Name of node, if not passed, project will be unloaded
                on all of the nodes.
            delete_sessions: If True, will delete all project sessions
                immediately before unloading.
        """
        on_nodes = self._normalize_nodes(on_nodes)
        all_nodes_are_selected = on_nodes is None or {
            node['name'] for node in self.nodes
        } == set(on_nodes)

        supports_server_wide_endpoint = meets_minimal_version(
            self.connection.iserver_version, '11.5.0500'
        )
        if delete_sessions is not None and not supports_server_wide_endpoint:
            logger.warning(
                "The `delete_sessions` argument requires iServer version 11.5.0500 or "
                f"later. Since your server version: {self.connection.iserver_version} "
                "is not compatible, this parameter will be omitted."
            )

        if all_nodes_are_selected and supports_server_wide_endpoint:
            self._change_project_state_all_nodes(
                status="unloaded", delete_sessions=delete_sessions
            )
        else:
            self.__change_project_state(func=self._unload_project, on_nodes=on_nodes)

    @method_version_handler('11.3.0800')
    def delete(self: Entity) -> bool:
        """Delete project.

        Returns:
            True if project was deleted, False otherwise.
        """
        self._DELETE_CONFIRM_MSG = (
            f"Are you sure you want to delete project "
            f"'{self.name}' with ID: {self._id}?\n"
            "All objects will be permanently deleted. This cannot be undone.\n"
            "Please type the project name to confirm: "
        )
        self._DELETE_SUCCESS_MSG = (
            f"Project '{self.name}' has been successfully deleted."
        )
        self._DELETE_PROMPT_ANSWER = self.name

        return super().delete(force=False)

    @method_version_handler('11.3.0000')
    def register(self, on_nodes: str | list | None = None) -> None:
        """Register project on nodes.

        A registered project will load on node (server) startup.

        Args:
            on_nodes: Name of node, if not passed, project will be loaded
                on all available nodes on startup.
        """
        if on_nodes is None:
            value = [node['name'] for node in self.nodes]
        else:
            on_nodes = on_nodes if isinstance(on_nodes, list) else [on_nodes]
            value = list(set(self.load_on_startup) | set(on_nodes))
        self._register(on_nodes=value)

    @method_version_handler('11.3.0000')
    def unregister(self, on_nodes: str | list | None = None) -> None:
        """Unregister project on nodes.

        An unregistered project will not load on node (server) startup.

        Args:
            on_nodes (str or list, optional): Name of node, if not passed,
                project will not be loaded on any nodes on startup.
        """
        if on_nodes is None:
            value = []
        else:
            on_nodes = on_nodes if isinstance(on_nodes, list) else [on_nodes]
            value = list(set(self.load_on_startup) - set(on_nodes))
        self._register(on_nodes=value)

    def update_settings(self) -> None:
        """Update the current project settings on the I-Server."""
        self.settings.update(self.id)

    def enable_caching(self) -> None:
        """Enable caching settings for the current project on the I-Server."""
        self.settings.enable_caching()
        self.update_settings()

    def disable_caching(self) -> None:
        """Disable caching settings for the current project on the I-Server."""
        self.settings.disable_caching()
        self.update_settings()

    @method_version_handler('11.3.0800')
    def delete_object_cache(self) -> None:
        """Delete object cache for the current project on the I-Server."""
        monitors.delete_caches(self.connection, self.id, 'object')
        if config.verbose:
            logger.info(
                f"Object cache was successfully deleted for project '{self.id}'"
            )

    @method_version_handler('11.3.0800')
    def delete_element_cache(self) -> None:
        """Delete element cache for the current project on the I-Server."""
        monitors.delete_caches(self.connection, self.id, 'element')
        if config.verbose:
            logger.info(
                f"Element cache was successfully deleted for project '{self.id}'"
            )

    def fetch_settings(self) -> None:
        """Fetch the current project settings from the I-Server."""
        self.settings.fetch(self.id)

    def list_caching_settings(self) -> dict:
        """
        Fetch current project settings connected with caching from I-Server
        """
        return self.settings.list_caching_properties()

    def is_loaded(self) -> bool:
        """Check if the project is loaded on any node (server)."""
        loaded = False
        self.fetch('nodes')
        if not isinstance(self.nodes, list):
            helper.exception_handler(
                "Could not retrieve current project status.",
                exception_type=ConnectionError,
            )
        for node in self.nodes:
            projects = node.get('projects')
            if projects:
                status = projects[0].get('status')
                loaded = status == 'loaded'
                if loaded:
                    break
        return loaded

    def get_data_engine_versions(self) -> dict:
        """Fetch the currently available data engine versions for project."""

        return projects.get_engine_settings(self.connection, self.id).json()['engine'][
            'versions'
        ]

    def get_current_engine_version(self):
        res = objects.get_property_set(
            self.connection,
            id=self.id,
            obj_type=self._OBJECT_TYPE.value,
            property_set_id=SYSTEM_OBJECTS_VERSION_PROPERTY_SET_ID,
        ).json()
        current_in_list = [
            prop for prop in res if prop['id'] == AE_VERSION_PROPERTY_INDEX
        ]
        current = current_in_list[0]['value'] if current_in_list else None

        return current

    def update_data_engine_version(self, new_version: str) -> None:
        """Update data engine version for project."""

        body = [
            {
                "properties": [{"value": new_version, "id": AE_VERSION_PROPERTY_INDEX}],
                "id": SYSTEM_OBJECTS_VERSION_PROPERTY_SET_ID,
            }
        ]
        objects.update_property_set(
            self.connection, id=self.id, obj_type=self._OBJECT_TYPE.value, body=body
        )

    def _register(self, on_nodes: list) -> None:
        path = f"/projects/{self.id}/nodes"
        body = {"operationList": [{"op": "replace", "path": path, "value": on_nodes}]}
        projects.update_projects_on_startup(self.connection, body)
        if config.verbose:
            if on_nodes:
                logger.info(f'Project will load on startup of: {on_nodes}')
            else:
                logger.warning('Project will not load on startup.')

    @method_version_handler('11.3.0100')
    def add_datasource(
        self, datasources: list['DatasourceInstance | str']
    ) -> list['DatasourceInstance']:
        """Add datasources to the project.

        Args:
            datasources (list["DatasourceInstance", str]): List of datasource
                objects or datasource IDs to be added to the project.

        Returns:
            List of datasource objects for the project.
        """
        from mstrio.datasources import DatasourceInstance, list_datasource_instances

        operation_list = []
        valid_datasource_ids = [
            datasource.id for datasource in list_datasource_instances(self.connection)
        ]
        datasource_ids = [
            datasource.id if isinstance(datasource, DatasourceInstance) else datasource
            for datasource in datasources
        ]

        for datasource_id in datasource_ids:
            if datasource_id not in valid_datasource_ids:
                raise ValueError(f"Datasource with ID {datasource_id} doesn't exist.")

            operation_list.append(
                {
                    'op': 'add',
                    'path': '/id',
                    'value': datasource_id,
                }
            )

        response = datasources_processors.update_project_datasources(
            connection=self.connection,
            id=self.id,
            body={'operationList': operation_list},
        )

        if config.verbose:
            logger.info(
                f"Datasources were successfully added to the project {self.id}."
            )

        return [
            DatasourceInstance.from_dict(datasource, self.connection)
            for datasource in response
        ]

    @method_version_handler('11.3.0100')
    def remove_datasource(self, datasources: list['DatasourceInstance | str']) -> None:
        """Remove datasources from the project.

        Args:
            datasources (list["DatasourceInstance", str]): List of datasource
                objects or datasource IDs to be removed from the project.
        """
        from mstrio.datasources import DatasourceInstance, list_datasource_instances

        operation_list = []
        valid_datasource_ids = [
            datasource.id
            for datasource in list_datasource_instances(
                self.connection, project=self.id
            )
        ]
        datasource_ids = [
            datasource.id if isinstance(datasource, DatasourceInstance) else datasource
            for datasource in datasources
        ]

        for datasource_id in datasource_ids:
            if datasource_id not in valid_datasource_ids:
                raise ValueError(
                    f"Datasource with ID {datasource_id} does not exist in the "
                    f"project {self.id}."
                )

            operation_list.append(
                {
                    'op': 'remove',
                    'path': '/id',
                    'value': datasource_id,
                }
            )

        datasources_api.update_project_datasources(
            connection=self.connection,
            id=self.id,
            body={'operationList': operation_list},
        )

        if config.verbose:
            logger.info(
                f"Datasources were successfully removed from the project {self.id}."
            )

    @method_version_handler('11.3.0600')
    def lock(self, lock_type: str | LockType, lock_id: str | None = None) -> None:
        """Lock the project.

        Args:
            lock_type (str, LockType): Lock type.
            lock_id (str, optional): Lock ID.
        """
        self.fetch('lock_status')

        if self.lock_status.lock_type != LockType.NOT_LOCKED:
            msg = (
                f"Project '{self.id}' is already locked with the lock type "
                f"`{self.lock_status.lock_type}`. "
                f"Please unlock it before applying a new lock."
            )
            raise ValueError(msg)

        lock_type = get_enum_val(lock_type, LockType)

        projects.update_project_lock(
            self.connection, self.id, {'lockType': lock_type, 'lockId': lock_id}
        )

        if config.verbose:
            logger.info(f"Project '{self.id}' locked.")

    @method_version_handler('11.3.0600')
    def unlock(
        self,
        lock_type: str | LockType | None = None,
        lock_id: str | None = None,
        force: bool = False,
    ) -> None:
        """Unlock the project.

        Args:
            lock_type (str, LockType, optional): Lock type.  Optional only if
                `force` is set to True.
            lock_id (str, optional): Lock ID.
            force (bool, optional): Whether to force unlock the project.
        """
        self.fetch('lock_status')

        if self.lock_status.lock_type == LockType.NOT_LOCKED:
            msg = f"Project '{self.id}' is not locked."
            raise ValueError(msg)

        if not force and not (lock_id and lock_type):
            msg = (
                "`lock_id` and `lock_type` must be provided to unlock the project "
                "when `force` is False."
            )
            raise ValueError(msg)

        if force and not lock_type:
            lock_type = self.lock_status.lock_type

        lock_type = get_enum_val(lock_type, LockType)

        projects.delete_project_lock(
            self.connection,
            self.id,
            lock_type=lock_type,
            lock_id=lock_id,
            force=force,
        )

        if config.verbose:
            logger.info(f"Project '{self.id}' unlocked.")

    def get_status_on_node(self, node: 'str | Node') -> IdleMode:
        """Get status of the project of specific node in the connected
        Intelligence server cluster.

        Args:
            node (str, Node): Name of the node or Node object.

        Returns:
            IdleMode: Status of the project on the node.
        """
        node_name = node if isinstance(node, str) else node.name
        return IdleMode(
            monitors.get_project_status_on_node(
                self.connection, node_name, self.id
            ).json()['status']
        )

    @method_version_handler('11.4.1200')
    def delete_unused_managed_objects(
        self,
        try_force_delete: bool = False,
        return_failed_items: bool = False,
    ) -> bool | list[dict]:
        """Delete all unused managed objects in the project.

        Managed Objects in scope of this method are objects of type:
        - ObjectTypes.ATTRIBUTE,
        - ObjectTypes.METRIC,
        - ObjectTypes.CONSOLIDATION,
        - ObjectTypes.CONSOLIDATION_ELEMENT,

        Note:
            This method is known to be resource and time intensive. Use only
            if necessary.

        Args:
            try_force_delete (bool, optional): If `True`, will attempt to delete
                items that could not be confirmed whether they are unused, for
                any reason. Server should not allow deleting of those and throw
                error which will be caught by this method. Defaults to `False`.
            return_failed_items (bool, optional): If `True`, will return a list
                of dicts of data of objects that could not be deleted. If
                `False`, will return a boolean indicating whether all unused
                objects were deleted successfully. Defaults to `False`.

        Returns:
            If `return_failed_items` is `False`, returns `True` if all unused
            managed objects were deleted successfully, `False` otherwise.
            If `return_failed_items` is `True`, returns a list of dicts of
            objects that could not be deleted. If all objects were deleted
            successfully, returns an empty list.
        """
        from mstrio.object_management.object import Object
        from mstrio.object_management.search_enums import SearchDomain, SearchScope
        from mstrio.object_management.search_operations import full_search

        items = full_search(
            self.connection,
            scope=SearchScope.MANAGED_ONLY,
            domain=SearchDomain.PROJECT,
            object_types=[
                ObjectTypes.ATTRIBUTE,
                ObjectTypes.METRIC,
                ObjectTypes.CONSOLIDATION,
                ObjectTypes.CONSOLIDATION_ELEMENT,
            ],
            project=self,
            include_hidden=True,
            to_dictionary=True,
        )

        if config.verbose:
            logger.info(f"Found {len(items)} managed objects to check if unused.")

        problematic_items: list[dict] = []
        final_items: list["Object"] = []
        final_len = 0  # cumulative final items length

        def bulk_delete(objs: list["Object"]) -> bool:
            FAIL_MSG = "Deleting some of the unused managed objects failed."

            try:
                res: Response = objects.bulk_delete_objects(
                    self.connection,
                    [obj.id for obj in objs],
                    [obj.type.value for obj in objs],
                    project_id=self.id,
                    error_msg=FAIL_MSG,
                )

                if config.verbose and res.ok:
                    logger.info(
                        "Successfully deleted a batch of unused managed "
                        f"objects in project '{self.id}'."
                    )

                return res.ok
            except Exception:
                logger.warning(FAIL_MSG)
                return False

        def perform_bulk_delete(checked: int, unused: int) -> None:
            nonlocal problematic_items, final_items

            if config.verbose:
                logger.info(
                    f"Checked {checked} items so far. Found {unused} are "
                    "confirmed unused."
                    + (" Deleting this new batch now." if final_items else "")
                )

            if final_items:
                if not bulk_delete(final_items):
                    # bulk delete may have failed due to only some of items,
                    # not all. Retry one by one to find problematic ones.
                    for obj in final_items:
                        try:
                            obj.delete(force=True)
                        except Exception:
                            problematic_items.append(obj.to_dict())

                final_items = []

        for i, itm in enumerate(items):
            if i % 100 == 0 and i > 0:
                perform_bulk_delete(i, final_len)

            try:
                obj = Object.from_dict(
                    itm,
                    self.connection,
                    with_missing_value=True,
                )

                if not obj.has_dependents():
                    final_len += 1
                    final_items.append(obj)
            except Exception:
                problematic_items.append(itm)

        perform_bulk_delete(len(items), final_len)

        if problematic_items and try_force_delete:
            # At this point we were either not able to delete the item, not
            # able to check if there are dependents, or something happened
            # during the bulk delete. As a last resort, we will try the most
            # raw delete method just to see.
            #
            # This will fail if there are dependents and this is what we want.
            if config.verbose:
                logger.info(
                    "Some items could not be confirmed whether they are unused. "
                    "Will try to delete them (this will fail if they are used)."
                )

            for prob_dict in problematic_items.copy():
                try:
                    with config.temp_verbose_disable():
                        objects.delete_object(
                            self.connection,
                            prob_dict['id'],
                            prob_dict['type'],
                            project_id=self.id,
                        )
                    problematic_items.remove(prob_dict)
                except Exception:
                    continue

        if problematic_items:
            logger.warning(
                "For the following items, could not check if they are unused "
                "or could not delete them: "
                f"{[p.get('id') for p in problematic_items]}. They were skipped."
            )

        if return_failed_items:
            return problematic_items
        return len(problematic_items) == 0

    @method_version_handler('11.5.0700')
    def duplicate(
        self,
        target_name: str,
        duplication_config: 'dict | DuplicationConfig | None' = None,
    ) -> 'ProjectDuplication':
        """Duplicate the project with a new name.

        Args:
            target_name (str): New name for the duplicated project.
            duplication_config (dict, DuplicationConfig, optional):
                Configuration for the duplication process. If not provided
                `DuplicationConfig()` with default values will be used.

        Returns:
            ProjectDuplication object.
        """

        if not duplication_config:
            duplication_config = DuplicationConfig()
        if isinstance(duplication_config, DuplicationConfig):
            duplication_config = duplication_config.to_dict()
        body = {
            "source": {
                "environment": {
                    "id": self.connection.base_url,
                    "name": self.connection.base_url,
                },
                "project": {
                    "id": self.id,
                    "name": self.name,
                },
            },
            "target": {
                "environment": {
                    "id": self.connection.base_url,
                    "name": self.connection.base_url,
                },
                "project": {"name": target_name},
            },
            "settings": duplication_config,
        }
        resp = projects.trigger_project_duplication(self.connection, self.id, body)
        return ProjectDuplication.from_dict(
            helper.get_response_json(resp), connection=self.connection
        )

    @property
    def load_on_startup(self):
        """View nodes (servers) to load project on startup."""
        response = projects.get_projects_on_startup(self.connection).json()
        return response['projects'][self.id]['nodes']

    @property
    def settings(self) -> "ProjectSettings":
        """`Settings` object storing Project settings.

        Settings can be listed by using `list_properties()` method.
        Settings can be modified directly by setting the values in the
        object.
        """

        if not hasattr(self, "_settings"):
            super(Entity, self).__setattr__(
                "_settings", ProjectSettings(self.connection, self.id)
            )
        return self._settings

    @settings.setter
    def settings(self, settings: "ProjectSettings") -> None:
        super(Entity, self).__setattr__("_settings", settings)

    @property
    def status(self):
        return self._status

    @property
    def alias(self):
        return self._alias

    @property
    def nodes(self):
        return self._nodes

    @property
    @method_version_handler(version='11.3.1200')
    def data_language_settings(self) -> DataLanguageSettings:
        if not hasattr(self._data_language_settings, '_connection'):
            self._data_language_settings._connection = self.connection
        if not hasattr(self._data_language_settings, '_project_id'):
            self._data_language_settings._project_id = self.id
        return self._data_language_settings

    @property
    @method_version_handler(version='11.3.1200')
    def metadata_language_settings(self) -> MetadataLanguageSettings:
        if not hasattr(self._metadata_language_settings, '_connection'):
            self._metadata_language_settings._connection = self.connection
        if not hasattr(self._metadata_language_settings, '_project_id'):
            self._metadata_language_settings._project_id = self.id
        return self._metadata_language_settings

    @property
    @method_version_handler('11.3.0600')
    def lock_status(self) -> LockStatus:
        return self._lock_status


class ProjectSettings(BaseSettings):
    """Object representation of Strategy One Project Settings.

    Used to fetch, view, modify, update, export to file, import from file and
    validate Project settings.

    The object can be optionally initialized with `connection` and
    `project_id`, which will automatically fetch the current settings for
    the specified project. If not specified, settings can be loaded from
    file using `import_from()` method. Object attributes (representing settings)
    can be modified manually. Lastly, the object can be applied to any
    project using the `update()` method.

    Attributes:
        connection: A Strategy One connection object
    """

    _TYPE = "allProjectSettings"
    _CONVERSION_MAP = {
        'maxCubeSizeForDownload': 'B',
        'maxDataUploadSize': 'B',
        'maxMstrFileSize': 'B',
        'maxHistoryListInboxMessage': 'B',
        'maxMemoryPerDataFetch': 'B',
        'maxElementCacheMemoryConsumption': 'B',
        'maxObjectCacheMemoryConsumption': 'B',
        'maxRWDCacheMemoryConsumption': 'B',
        'maxReportCacheMemoryConsumption': 'B',
        'maxCubeQuota': 'B',
        'maxCubeMemUsage': 'B',
        'maxSqlGenerationMemoryConsumption': 'B',
        'reportCacheLifeTime': 'hour',
        'maxWarehouseJobExecTime': 'sec',
        'maxReportExecutionTime': 'sec',
        'maxScheduledReportExecutionTime': 'sec',
        'statisticsPurgeTimeout': 'sec',
        'maxPromptWaitingTime': 'sec',
        'maxRAMForReportRWDCacheIndex': '%',
        'cubeIndexGrowthUpperBound': '%',
    }
    _CACHING_SETTINGS_TO_ENABLE = (
        "enableReportServerCaching",
        "enableCachingForPromptedReportDocument",
        "enableCachingForNonPromptedReportDocument",
        "enableXmlCachingForReport",
    )
    _CACHING_SETTINGS_TO_DISABLE = _CACHING_SETTINGS_TO_ENABLE + (
        "recordPromptAnswersForCacheMonitoring",
        "enableDocumentOutputCachingInXml",
        "enableDocumentOutputCachingInHtml",
        "enableDocumentOutputCachingInPdf",
        "enableDocumentOutputCachingInExcel",
    )

    def __init__(self, connection: Connection, project_id: str | None = None):
        """Initialize `ProjectSettings` object.

        Args:
            connection: Strategy One connection object returned by
                `connection.Connection()`.
            project_id: Project ID
        """
        super(BaseSettings, self).__setattr__('_connection', connection)
        super(BaseSettings, self).__setattr__('_project_id', project_id)
        self._configure_settings()

        if project_id:
            self.fetch()

    def fetch(self, project_id: str | None = None) -> None:
        """Fetch current project settings from I-Server and update this
        `ProjectSettings` object.

        Args:
            project_id: Project ID
        """
        self._check_params(project_id)
        super().fetch()

    def update(self, project_id: str | None = None) -> None:
        """Update the current project settings on I-Server using this
        Settings object.

        Args:
            project_id: Project ID
        """
        self._check_params(project_id)
        set_dict = self._prepare_settings_push()
        if not set_dict and config.verbose:
            logger.info('No settings to update.')
            return
        response = projects.update_project_settings(
            self._connection, self._project_id, set_dict
        )
        if config.verbose:
            if response.status_code == 200:
                logger.info('Project settings updated.')
            elif response.status_code == 207:
                partial_succ = response.json()
                logging.info(f'Project settings partially successful.\n{partial_succ}')
        super().update()

    def __setattr__(self, key, value) -> None:
        if isinstance(value, Enum):
            value = get_enum_val(value, type(value))

        if isinstance(value, list):
            value = [
                get_enum_val(v, type(v)) if isinstance(v, Enum) else v for v in value
            ]

        super().__setattr__(key, value)

    @wip()
    def enable_caching(self) -> None:
        """
        Enable caching settings for the current project on I-Server
        using this Settings object
        """
        for setting in self._CACHING_SETTINGS_TO_ENABLE:
            if hasattr(self, setting):
                self.__setattr__(setting, True)

    @wip()
    def disable_caching(self) -> None:
        """
        Disable caching settings for the current project on I-Server
        using this Settings object
        """
        for setting in self._CACHING_SETTINGS_TO_DISABLE:
            if hasattr(self, setting):
                self.__setattr__(setting, False)

    def _fetch(self) -> dict:
        response = projects.get_project_settings(
            self._connection, self._project_id, whitelist=[('ERR001', 404)]
        )
        settings = response.json() if response.ok else {}
        if not response.ok:
            msg = (
                "Settings could not be fetched. It may be because the project is not "
                "loaded in the Intelligence Server or the project is idle."
            )
            helper.exception_handler(msg, Warning)
        return self._prepare_settings_fetch(settings)

    def _get_config(self):
        if not ProjectSettings._CONFIG:
            project_id = self._project_id
            if not project_id:
                project_id = Project._list_loaded_projects(
                    self._connection, to_dictionary=True
                )['id'][0]
            response = projects.get_project_settings_config(
                self._connection, project_id
            )
            ProjectSettings._CONFIG = response.json()
            super()._get_config()

    def _check_params(self, project_id: str | None = None):
        if project_id:
            super(BaseSettings, self).__setattr__('_project_id', project_id)
        if not self._connection or not self._project_id:
            raise AttributeError(
                "Please provide `connection` and `project_id` parameter"
            )

    @wip()
    def list_caching_properties(self, show_description: bool = False) -> dict:
        """
        Fetch current project settings connected with caching from I-Server

        Args:
            show_description (bool): if True return, description and value
                for each setting, else, return values only
        """
        self.fetch()
        return {
            k: v
            for (k, v) in self.list_properties(
                show_description=show_description
            ).items()
            if any(word in k.lower() for word in ("cache", "caching"))
        }


@dataclass
class DuplicationConfig(helper.Dictable):
    """Configuration for project duplication.

    Attributes:
        schema_objects_only (bool, optional): If True, only schema objects will
            be duplicated.
        skip_empty_profile_folders (bool, optional): Whether to skip empty
            profile folders during duplication.
        include_user_subscriptions (bool, optional): Whether to include user
            subscriptions in the duplication.
        include_contact_subscriptions (bool, optional): Whether to include
            contact subscriptions in the duplication.
        import_description (str, optional): Description for the import operation
            to be stored in ProjectDuplication class object.
        import_default_locale (int, optional): Locale used for
            internationalization in imported project. Please provide Language
            object 'lcid' attribute.
        import_locales (list[int], optional): List of locale ids for imported
            project. Please provide Language object 'lcid' attribute.
    """

    schema_objects_only: bool = True
    skip_empty_profile_folders: bool = True
    include_user_subscriptions: bool = True
    include_contact_subscriptions: bool = True
    import_description: str | None = 'Project Duplication'
    import_default_locale: int | None = 0
    import_locales: list[int] | None = field(default_factory=lambda: [0])

    def __post_init__(self):
        if self.import_default_locale not in self.import_locales:
            raise ValueError("Default locale must be included in the import locales.")

    def to_dict(self) -> dict:
        """Convert the duplication configuration to a dictionary accepted by
        the API."""
        settings = {
            "export": {
                "projectObjectsPreference": {
                    "schemaObjectsOnly": self.schema_objects_only,
                    "skipEmptyProfileFolders": self.skip_empty_profile_folders,
                },
                "subscriptionPreferences": {
                    "includeUserSubscriptions": self.include_user_subscriptions,
                    "includeContactSubscriptions": self.include_contact_subscriptions,
                },
            },
            "import": {
                "description": self.import_description,
                "defaultLocale": self.import_default_locale,
                "locales": self.import_locales,
            },
        }
        return settings

    @classmethod
    def from_dict(cls, source, **kwargs):
        """Initialize DuplicationConfig object from raw settings dictionary."""
        source = helper.camel_to_snake(source)
        export_sets = source.get('export', {})
        import_sets = source.get('import', {})
        obj = DuplicationConfig(
            schema_objects_only=export_sets.get('project_objects_preference', {}).get(
                'schema_objects_only', True
            ),
            skip_empty_profile_folders=export_sets.get(
                'project_objects_preference', {}
            ).get('skip_empty_profile_folders', True),
            include_user_subscriptions=export_sets.get(
                'subscription_preferences', {}
            ).get('include_user_subscriptions', True),
            include_contact_subscriptions=export_sets.get(
                'subscription_preferences', {}
            ).get('include_contact_subscriptions', True),
            import_description=import_sets.get('description'),
            import_default_locale=import_sets.get('default_locale', 0),
            import_locales=import_sets.get('locales', [0]),
        )
        return obj


@dataclass
class ProjectInfo(helper.Dictable):
    """Dataclass to store project duplication record info. It stores
    environment, project, and creator info."""

    creator: dict = field(default_factory=dict)
    environment: dict = field(default_factory=dict)
    project: dict = field(default_factory=dict)

    @classmethod
    def from_dict(cls, source, **kwargs):
        """Initialize ProjectInfo from a dictionary."""
        return cls(
            creator=source.get('creator', {}),
            environment=source.get('environment', {}),
            project=source.get('project', {}),
        )


class ProjectDuplicationStatus(AutoUpperName):
    """Enum representing the status of a project duplication."""

    UNKNOWN = auto()
    CREATED = auto()
    CREATE_FAILED = auto()
    EXPORTING = auto()
    EXPORTED = auto()
    EXPORT_FAILED = auto()
    IMPORTING = auto()
    IMPORT_FAILED = auto()
    COMPLETED = auto()
    CANCELLED = auto()
    CANCELLING = auto()
    EXPORT_SYNCING = auto()
    IMPORT_SYNCING = auto()


@method_version_handler('11.5.0700')
def list_projects_duplications(
    connection: Connection, limit: int | None = None, to_dictionary: bool = False
) -> list['ProjectDuplication | dict']:
    """Get project duplications.

    Args:
        connection: Strategy One connection object returned by
            `connection.Connection()`
        limit (int, optional): Limit the number of elements returned. If `None`
            (default), all objects are returned.
        to_dictionary (bool, optional): If `True` returns dict, by default
            (False) returns ProjectDuplication objects.

    Returns:
        List of ProjectDuplication objects or list od dicts.
    """
    duplications = projects_processors.get_project_duplications(
        connection=connection, limit=limit
    )

    duplications = helper.camel_to_snake(duplications)
    if to_dictionary:
        return duplications
    return [
        ProjectDuplication.from_dict(
            source=duplication, connection=connection, with_missing_value=False
        )
        for duplication in duplications
    ]


class ProjectDuplication(EntityBase):
    """Object representation of a project duplication request.

    Attributes:
        id (str): Duplication job ID.
        source (ProjectInfo): Source environment, project, and creator info.
        target (ProjectInfo): Target environment, project, and creator info.
        created_date (datetime): Duplication creation date.
        last_updated_date (datetime): Duplication last update date.
        status (ProjectDuplicationStatus): Duplication status.
        progress (int): Duplication progress percentage.
        message (str): Status or error message.
        settings (DuplicationConfig): Duplication settings.
    """

    _API_GETTERS = {
        (
            'id',
            'source',
            'target',
            'created_date',
            'last_updated_date',
            'status',
            'progress',
            'message',
            'settings',
        ): projects.get_project_duplication
    }

    _FROM_DICT_MAP = {
        **EntityBase._FROM_DICT_MAP,
        'created_date': DatetimeFormats.FULLDATETIME,
        'last_updated_date': DatetimeFormats.FULLDATETIME,
        'settings': DuplicationConfig.from_dict,
        'source': ProjectInfo.from_dict,
        'status': ProjectDuplicationStatus,
        'target': ProjectInfo.from_dict,
    }

    def __init__(
        self,
        connection: "Connection",
        id: str,
    ) -> None:
        """Initialize ProjectDuplication object.

        Args:
            connection: Strategy One connection object returned by
                `connection.Connection()`.
            id: ID of the ProjectDuplication
        """

        super().__init__(connection=connection, object_id=id)

    def _init_variables(self, **kwargs) -> None:
        super()._init_variables(**kwargs)
        self.source = (
            ProjectInfo.from_dict(kwargs.get('source', {}))
            if kwargs.get('source')
            else {}
        )
        self.target = (
            ProjectInfo.from_dict(kwargs.get('target', {}))
            if kwargs.get('target')
            else {}
        )
        self.created_date = time_helper.map_str_to_datetime(
            "created_date", kwargs.get("created_date"), self._FROM_DICT_MAP
        )
        self.last_updated_date = time_helper.map_str_to_datetime(
            "last_updated_date", kwargs.get("last_updated_date"), self._FROM_DICT_MAP
        )
        self.status = (
            ProjectDuplicationStatus(kwargs.get('status'))
            if kwargs.get('status')
            else None
        )
        self.progress = kwargs.get('progress', 0)
        self.message = kwargs.get('message', '')
        self.settings = (
            DuplicationConfig.from_dict(kwargs.get('settings'))
            if kwargs.get('settings')
            else None
        )

    @classmethod
    def from_dict(
        cls,
        source: dict,
        connection: 'Connection | None' = None,
        with_missing_value: bool = False,
    ):
        return super().from_dict(
            source=source,
            connection=connection,
            with_missing_value=with_missing_value,
        )

    def cancel(self) -> bool:
        """Cancel the project duplication job.

        Returns:
            bool: True if the cancellation was successful, False otherwise.
        """
        body = {"status": "cancelled"}
        response = projects.cancel_project_duplication(self.connection, self.id, body)
        if response.status_code == 204:
            if config.verbose:
                logger.info(
                    f"Project duplication with ID '{self.id}' cancelled successfully."
                )
            return True
        else:
            if config.verbose:
                logger.error(
                    f"Failed to cancel project duplication with ID '{self.id}'."
                )
            return False

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(id='{self.id}', "
            f"progress={self.progress}, "
            f"source_project='{self.source.project.get('name')}', "
            f"target_project='{self.target.project.get('name')}')"
        )

    def wait_for_stable_status(
        self, timeout: int = 240, interval: int = 2
    ) -> 'ProjectDuplication':
        """Wait for the project duplication to reach a stable status.

        Args:
            timeout (int, optional): Maximum time to wait in seconds.
            interval (int, optional): Time between status checks in seconds.

        Returns:
            ProjectDuplication: The project duplication object with updated
                status.
        """
        not_stable_statuses = [
            ProjectDuplicationStatus.CANCELLING,
            ProjectDuplicationStatus.EXPORTING,
            ProjectDuplicationStatus.EXPORT_SYNCING,
            ProjectDuplicationStatus.IMPORTING,
            ProjectDuplicationStatus.IMPORT_SYNCING,
        ]
        return helper.wait_for_stable_status(
            self, 'status', not_stable_statuses, timeout=timeout, interval=interval
        )



from enum import Enum


class AutoName(Enum):
    def _generate_next_value_(self, start, count, last_values):
        return self.lower()

    @classmethod
    def has_value(cls, value):
        return value in {item.value for item in cls}

    def __repr__(self) -> str:
        return self.__str__()


class AutoUpperName(Enum):
    def _generate_next_value_(self, start, count, last_values):
        return self


class AutoCapitalizedName(Enum):
    def _generate_next_value_(self, start, count, last_values):
        return self.capitalize()


def __get_enum_helper(
    obj,
    enum: type[Enum] = Enum,
    get_value: bool = False,
    throw_on_unknown: bool = True,
) -> str | int | type[Enum] | None:
    """Helper function for `get_enum` and `get_enum_val`."""
    if obj is None:
        return obj

    if isinstance(obj, enum):
        return obj.value if get_value else obj

    if isinstance(obj, (str, int)):
        validate_enum_value(obj, enum, throw_on_unknown=throw_on_unknown)
        return obj if get_value else enum(obj)

    raise TypeError(f"Incorrect type. Value should be of type: {enum}.")


def get_enum(
    obj, enum: type[Enum] = Enum, throw_on_unknown: bool = True
) -> type[Enum] | None:
    """Safely get enum from enum or str."""
    return __get_enum_helper(obj, enum, throw_on_unknown=throw_on_unknown)


def get_enum_val(
    obj, enum: type[Enum] = Enum, throw_on_unknown: bool = True
) -> str | int | None:
    """Safely extract value from enum or str."""
    return __get_enum_helper(
        obj, enum, get_value=True, throw_on_unknown=throw_on_unknown
    )


def validate_enum_value(
    obj: str | int,
    enum: type[Enum] | tuple[type[Enum]] | list[type[Enum]],
    throw_on_unknown: bool = True,
) -> None:
    """Validate provided value. If not correct,
    error message with possible options will be displayed.
    """
    from mstrio.utils.helper import exception_handler

    possible_values = [
        e.value
        for item in enum
        for e in (item if isinstance(enum, (tuple, list)) else [item])
    ]
    err_msg = f"Unknown enum value '{obj}'. Known possible values are {possible_values}"

    if obj not in possible_values:
        exception_handler(
            err_msg, exception_type=ValueError if throw_on_unknown else Warning
        )



"""
Manage active and inactive users.
Perform administrative tasks such as setting default language, listing
subscriptions.

1. Connect to environment using data from workstation
2. Get language object based on provided id
3. For each active user set up language which was retrieved in step 2
   as the default one
4. Get list of all inactive users
5. Print number of inactive users
6. For each inactive user get list of its subscriptions and print them with
   user name
"""

from mstrio.distribution_services.subscription.subscription_manager import (
    list_subscriptions,
)
from mstrio.server.language import Language
from mstrio.users_and_groups.user import list_users
from mstrio.connection import get_connection

# connect to environment without providing user credentials
# variable `workstationData` is stored within Workstation
conn = get_connection(workstationData, project_name='MicroStrategy Tutorial')

ALTERED_LANGUAGE = Language(conn, id='000004124F95EF3956E52781700C1E7A')

# list active users
active_users = list_users(conn, enabled=True)

# set up default language for those users
for user in active_users:
    user.alter(language=ALTERED_LANGUAGE)

# list all inactive users and count them
inactive_users = list_users(conn, enabled=False)
inactive_users_count = len(inactive_users)
print('Number of inactive users:', inactive_users_count)

# list all related subscriptions for inactive users
for user in inactive_users:
    subscriptions = list_subscriptions(conn, project_id=conn.project_id, owner=user)
    print('Subscriptions for user', user.name, ':', subscriptions)



import logging
from enum import auto
from typing import TYPE_CHECKING

from mstrio import config
from mstrio.api import user_hierarchies
from mstrio.modeling.schema.helpers import SchemaObjectReference
from mstrio.types import ObjectTypes
from mstrio.users_and_groups import User
from mstrio.utils.entity import CopyMixin, DeleteMixin, Entity, MoveMixin
from mstrio.utils.enum_helper import AutoName
from mstrio.utils.helper import (
    Dictable,
    delete_none_values,
    fetch_objects,
    find_object_with_name,
    get_args_from_func,
    get_default_args_from_func,
    get_enum_val,
)
from mstrio.utils.response_processors import objects as objects_processors
from mstrio.utils.version_helper import class_version_handler, method_version_handler

if TYPE_CHECKING:
    from mstrio.connection import Connection

logger = logging.getLogger(__name__)


class ElementDisplayOption(AutoName):
    LIMITED_ELEMENTS = auto()
    ALL_ELEMENTS = auto()
    NO_ELEMENTS = auto()


class UserHierarchySubType(AutoName):
    DIMENSION_USER = auto()
    DIMENSION_USER_HIERARCHY = auto()


@method_version_handler('11.3.0200')
def list_user_hierarchies(
    connection: "Connection", to_dictionary: bool = False, limit: int = None, **filters
) -> list["UserHierarchy"] | list[dict]:
    """Get list of UserHierarchy objects or dicts. Optionally filter the
    user hierarchies by specifying filters.

    Args:
        connection: Strategy One connection object returned by
            `connection.Connection()`
        to_dictionary: If True returns dict, by default (False) returns
            User Hierarchy objects.
        limit: limit the number of elements returned. If `None` (default), all
            objects are returned.
        **filters: Available filter parameters: ['id', 'name', 'sub_type']

    Examples:
        >>> list_user_hierarchies(connection, name='hierarchy_name')
    """
    return UserHierarchy._list_user_hierarchies(
        connection=connection,
        to_dictionary=to_dictionary,
        limit=limit,
        **filters,
    )


class HierarchyAttribute(Dictable):
    """Object that specifies the hierarchy attribute.

    Attributes:
        object_id: hierarchy attribute ID
        name: name of the attribute
        entry_point: whether this hierarchy attribute is an entry point
        element_display_option: used to control element display for
            each hierarchy attribute, ElementDisplayOption enum
        filters: the list of filters defined on this hierarchy attribute
        limit: the element display limit when element_display_option
            is limited_elements
    """

    _FROM_DICT_MAP = {
        'element_display_option': ElementDisplayOption,
        'filters': [SchemaObjectReference.from_dict],
    }

    def __init__(
        self,
        object_id: str,
        entry_point: bool,
        name: str,
        element_display_option: ElementDisplayOption | str,
        filters: list[SchemaObjectReference] | list[dict] = None,
        limit: int | None = None,
    ):
        self.object_id = object_id
        self.entry_point = entry_point
        self.name = name
        self.element_display_option = (
            element_display_option
            if isinstance(element_display_option, ElementDisplayOption)
            else ElementDisplayOption(element_display_option)
        )
        self.filters = (
            [
                (
                    SchemaObjectReference.from_dict(obj_ref)
                    if not isinstance(obj_ref, SchemaObjectReference)
                    else object_id
                )
                for obj_ref in filters
            ]
            if filters
            else None
        )
        self.limit = limit

    def __eq__(self, other):
        if isinstance(self, dict):
            self = HierarchyAttribute.from_dict(self)
        if isinstance(other, dict):
            other = HierarchyAttribute.from_dict(other)
        return self.name == other.name and self.object_id == other.object_id


class HierarchyRelationship(Dictable):
    """Object that specifies the hierarchy relationship between
    hierarchy attributes.

    Attributes:
        parent: an information about an object representing a
            parent in a current hierarchy relationship,
            SchemaObjectReference object
        child: an information about an object representing a
            child in a current hierarchy relationship,
            SchemaObjectReference object
    """

    _FROM_DICT_MAP = {
        'parent': SchemaObjectReference.from_dict,
        'child': SchemaObjectReference.from_dict,
    }

    def __init__(
        self,
        parent: SchemaObjectReference | dict,
        child: SchemaObjectReference | dict,
    ):
        self.parent = (
            parent
            if isinstance(parent, SchemaObjectReference)
            else SchemaObjectReference.from_dict(parent)
        )
        self.child = (
            child
            if isinstance(child, SchemaObjectReference)
            else SchemaObjectReference.from_dict(child)
        )

    def __eq__(self, other):
        if isinstance(self, dict):
            self = HierarchyRelationship.from_dict(self)
        if isinstance(other, dict):
            other = HierarchyRelationship.from_dict(other)
        return self.parent == other.parent and self.child == other.child


@class_version_handler('11.3.0200')
class UserHierarchy(Entity, CopyMixin, MoveMixin, DeleteMixin):
    """A unique abstraction of hierarchies above the System Hierarchy,
    which can contain an arbitrary number of attributes and paths between
    them. These User Hierarchies allow users to browse through the data
    and drill as required by business needs, as opposed to how the data
    is physically stored in the data source.

    Attributes:
        name: name of the user hierarchy
        id: user hierarchy ID
        description: description of the user hierarchy
        sub_type: string literal used to identify the type of a metadata object,
            UserHierarchySubType enum
        version: object version ID
        ancestors: list of ancestor folders
        type: object type, ObjectTypes enum
        ext_type: object extended type, ExtendedType enum
        date_created: creation time, DateTime object
        date_modified: last modification time, DateTime object
        owner: User object that is the owner
        acg: access rights (See EnumDSSXMLAccessRightFlags for possible values)
        acl: object access control list
        version_id: the version number this object is currently carrying
        use_as_drill_hierarchy: whether this user hierarchy is used as
            drill hierarchy, default True
        is_embedded: if true indicates that the target object of this reference
            is embedded within this object, if this field is omitted
            (as is usual) then the target is not embedded.
        path: the path of the object, read only
        primary_locale: the primary locale of the object, in the IETF BCP 47
            language tag format, such as "en-US", read only
        attributes: the list of user hierarchy attributes
        relationships: the list of attribute relationships stored in
            the user hierarchy
        destination_folder_id: a globally unique identifier used to distinguish
            between metadata objects within the same project
        hidden: Specifies whether the object is hidden
    """

    _OBJECT_TYPE = ObjectTypes.DIMENSION
    _FROM_DICT_MAP = {
        **Entity._FROM_DICT_MAP,
        'owner': User.from_dict,
        'sub_type': UserHierarchySubType,
        'attributes': [HierarchyAttribute.from_dict],
        'relationships': [HierarchyRelationship.from_dict],
    }
    _API_GETTERS = {
        (
            'type',
            'subtype',
            'ext_type',
            'date_created',
            'date_modified',
            'version',
            'owner',
            'ancestors',
            'acg',
            'acl',
            'hidden',
            'comments',
        ): objects_processors.get_info,
        (
            'id',
            'name',
            'description',
            'sub_type',
            'date_created',
            'date_modified',
            'path',
            'version_id',
            'use_as_drill_hierarchy',
            'is_embedded',
            'primary_locale',
            'attributes',
            'relationships',
            'destination_folder_id',
        ): user_hierarchies.get_user_hierarchy,
    }
    _API_PATCH: dict = {
        (
            'name',
            'description',
            'relationships',
            'use_as_drill_hierarchy',
            'destination_folder_id',
            'sub_type',
            'is_embedded',
            'attributes',
        ): (user_hierarchies.update_user_hierarchy, "put"),
        (
            'folder_id',
            'hidden',
            'comments',
            'owner',
        ): (objects_processors.update, 'partial_put'),
    }
    _API_DELETE = staticmethod(user_hierarchies.delete_user_hierarchy)
    _PATCH_PATH_TYPES = {
        **Entity._PATCH_PATH_TYPES,
        'sub_type': str,
        'use_as_drill_hierarchy': bool,
        'destination_folder_id': str,
        'is_embedded': bool,
        'attributes': list,
        'relationships': list,
    }
    _REST_ATTR_MAP = {
        "object_id": "id",
    }

    def __init__(
        self,
        connection: "Connection",
        id: str | None = None,
        name: str | None = None,
    ):
        """Initialize user hierarchy object by its identifier.

        Note:
            Parameter `name` is not used when fetching. If only `name` parameter
            is provided, `id` will be found automatically if such object exists.

        Args:

        """
        if id is None:
            if name is None:
                raise ValueError(
                    "Please specify either 'name' or 'id' parameter in the constructor."
                )

            object_info, object_info["connection"] = (
                find_object_with_name(
                    connection=connection,
                    cls=self.__class__,
                    name=name,
                    listing_function=list_user_hierarchies,
                    search_pattern=None,
                ),
                connection,
            )
            self._init_variables(**object_info)
        else:
            super().__init__(connection, id)

    def _init_variables(self, **kwargs) -> None:
        super()._init_variables(**kwargs)
        self.sub_type = (
            UserHierarchySubType(kwargs.get('sub_type'))
            if kwargs.get('sub_type')
            else None
        )
        self.primary_locale = kwargs.get('primary_locale')
        self.is_embedded = kwargs.get('is_embedded')
        self.destination_folder_id = kwargs.get('destination_folder_id')
        self.use_as_drill_hierarchy = kwargs.get('use_as_drill_hierarchy')
        self.attributes = (
            [
                HierarchyAttribute.from_dict(source=attr)
                for attr in kwargs.get("attributes")
            ]
            if kwargs.get('attributes')
            else None
        )
        self.relationships = (
            [
                HierarchyRelationship.from_dict(source=rel)
                for rel in kwargs.get("relationships")
            ]
            if kwargs.get('attributes')
            else None
        )
        self._path = kwargs.get('path')
        self._version_id = kwargs.get('version_id')

    @classmethod
    def create(
        cls,
        connection: "Connection",
        name: str,
        sub_type: str | UserHierarchySubType,
        destination_folder_id: str,
        attributes: list[HierarchyAttribute] | list[dict],
        use_as_drill_hierarchy: bool = True,
        description: str = None,
        is_embedded: bool = False,
        primary_locale: str = None,
        relationships: list[HierarchyRelationship] | list[dict] = None,
    ):
        """Create a new user hierarchy in a specific project.

        Args:
            connection: Strategy One connection object returned by
                `connection.Connection()`.
            name (str): name of a new user hierarchy.
            sub_type (str, enum): string literal used to identify the type of
                a metadata object, UserHierarchySubType enum
            attributes (list): the list of attributes that
                do have any relationships currently
            destination_folder_id (str): a globally unique identifier used
                to distinguish between metadata objects within the same project
            use_as_drill_hierarchy (bool, optional): whether this user hierarchy
                is used as drill hierarchy
            description (str, optional): optional description of a new
                user hierarchy
            is_embedded (bool, optional): if true indicates that the target
                object of this reference is embedded within this object
            primary_locale (str, optional): the primary locale of the object,
                in the IETF BCP 47 language tag format, such as "en-US"
            relationships (list, optional): the list of attribute
                relationships stored in the system hierarchy

        Returns:
            UserHierarchy object
        """
        attributes = [
            attr.to_dict() if isinstance(attr, HierarchyAttribute) else attr
            for attr in attributes
        ]
        relationships = (
            [
                rel.to_dict() if isinstance(rel, HierarchyRelationship) else rel
                for rel in relationships
            ]
            if relationships
            else None
        )
        body = {
            "information": {
                "name": name,
                "description": description,
                "subType": get_enum_val(sub_type, UserHierarchySubType),
                "destinationFolderId": destination_folder_id,
                "isEmbedded": is_embedded,
                "primaryLocale": primary_locale,
            },
            "useAsDrillHierarchy": use_as_drill_hierarchy,
            "primaryLocale": primary_locale,
            "attributes": attributes,
            "relationships": relationships,
        }
        body = delete_none_values(body, recursion=True)
        response = user_hierarchies.create_user_hierarchy(connection, body).json()
        if config.verbose:
            logger.info(
                f"Successfully created user hierarchy named: '{response['name']}' "
                f"with ID: '{response['id']}'"
            )
        return cls.from_dict(source=response, connection=connection)

    def alter(
        self,
        name: str = None,
        sub_type: str | UserHierarchySubType = None,
        attributes: list[HierarchyAttribute] | list[dict] | None = None,
        use_as_drill_hierarchy: bool | None = None,
        description: str | None = None,
        is_embedded: bool = False,
        destination_folder_id: str | None = None,
        relationships: list[HierarchyRelationship] | list[dict] | None = None,
        hidden: bool | None = None,
        comments: str | None = None,
        owner: str | User | None = None,
    ):
        """Alter the user hierarchies properties.

        Args:
            name (str, optional): name of a user hierarchy
            sub_type (str, enum, optional):  string literal used to identify
                the type of a metadata object, UserHierarchySubType enum
            description(str, optional): description of a user hierarchy
            use_as_drill_hierarchy (bool, optional): whether this user hierarchy
                is used as drill hierarchy
            attributes (list, optional): the list of attributes that
                do have any relationships currently
            is_embedded (bool, optional): if true indicates that the target
                object of this reference is embedded within this object
            destination_folder_id (str, optional): a globally unique identifier
                used to distinguish between objects within the same project
            relationships (list, optional): the list of attribute
                relationships stored in the system hierarchy
            hidden (bool, optional): Specifies whether the object is hidden.
                Default value: False.
            comments (str, optional): long description of the user hierarchy
            owner: (str, User, optional): owner of the user hierarchy
        """
        if isinstance(owner, User):
            owner = owner.id
        func = self.alter
        args = get_args_from_func(func)
        defaults = get_default_args_from_func(func)
        default_dict = dict(zip(args[-len(defaults) :], defaults)) if defaults else {}
        local = locals()
        properties = {}
        for property_key in default_dict.keys():
            if local[property_key] is not None:
                properties[property_key] = local[property_key]
        self._alter_properties(**properties)

    def add_attribute(self, attribute: HierarchyAttribute | dict):
        """Add an attribute to an existing hierarchy.

        Args:
            attribute: (HierarchyAttribute, dict): an attribute
                to be added to the hierarchy
        """
        return self.alter(attributes=self.attributes + [attribute])

    def remove_attribute(self, attribute: HierarchyAttribute | dict):
        """Remove attribute from an existing user hierarchy.

        Args:
            attribute: (HierarchyAttribute, dict): an attribute
                to be removed from the hierarchy
        """
        attributes = [attr for attr in self.attributes if attr != attribute]
        return self.alter(attributes=attributes)

    def add_relationship(self, relationship: HierarchyRelationship | dict):
        """Add a relationship to an existing hierarchy.

        Args:
            relationship: (HierarchyRelationship, dict): a relationship
                to be added to the hierarchy
        """
        return self.alter(relationships=self.relationships + [relationship])

    def remove_relationship(self, relationship: HierarchyRelationship | dict):
        """Remove relationship from an existing user hierarchy.

        Args:
            relationship: (HierarchyRelationship, dict): a relationship
                to be removed from the hierarchy
        """
        relationships = [
            rel.to_dict() for rel in self.relationships if rel != relationship
        ]
        body = self.to_dict()
        body['relationships'] = relationships
        response = user_hierarchies.update_user_hierarchy(
            connection=self.connection, id=self.id, body=body
        )
        if response.ok:
            response = response.json()
            self._set_object_attributes(**response)

    @classmethod
    def _list_user_hierarchies(
        cls,
        connection: "Connection",
        to_dictionary: bool = False,
        limit: int = None,
        **filters,
    ) -> list["UserHierarchy"] | list[dict]:
        objects = fetch_objects(
            connection=connection,
            api=user_hierarchies.get_user_hierarchies,
            dict_unpack_value="hierarchies",
            limit=limit,
            filters=filters,
        )
        if to_dictionary:
            return objects
        return [cls.from_dict(source=obj, connection=connection) for obj in objects]

    @property
    def path(self):
        return self._path

    @property
    def version_id(self):
        return self._version_id



"""This is the demo script to show how to manage Search Objects. Its basic goal
is to present what can be done with the SearchObject class, its related methods
and to ease its usage.

This script will not work without replacing parameters with real values.
"""

from mstrio.connection import get_connection
from mstrio.object_management.folder import Folder
from mstrio.object_management.search_operations import (
    DateQuery,
    SearchObject,
    list_search_objects,
)
from mstrio.types import ObjectTypes

# Define variables which can be later used in a script
PROJECT_NAME = $project_name  # Project to connect to
FOLDER_ID = $folder_id  # ID of the folder where the Search Object will be created
FOLDER_QUERY_LISTING = $folder_query_listing  # ID of the folder where the Search Object listing will be narrowed down to
NAME_QUERY_1 = $name_query_1  # Object name to search for
NAME_QUERY_2 = $name_query_2  # Object name to search for (after update)
FOLDER_QUERY_1 = $folder_query_1  # Root folder to search in
FOLDER_QUERY_2 = $folder_query_2  # Root folder to search in (after update)


# Connect to the Strategy One environment
conn = get_connection(workstationData, project_name=PROJECT_NAME)

# List all Search Objects
search_objects = list_search_objects(conn)
print(f"Found {len(search_objects)} search objects.")

# List Search Objects in a given folder
search_objects = list_search_objects(conn, folder_id=FOLDER_QUERY_LISTING)
print(f"Found {len(search_objects)} search objects in folder.")

# Create a Search Object
new_so = SearchObject.create(
    conn,
    name='New Search Object',
    destination_folder=FOLDER_ID,
    project_name=PROJECT_NAME,
    name_query=NAME_QUERY_1,
    root_folder_query=FOLDER_QUERY_1,
    include_subfolders=True,
)

# List properties of the Search Object
properties = new_so.list_properties()
print(properties)

# Run search based on parameters stored in the Search Object
results = new_so.run(project_name=PROJECT_NAME)
print(f"Search returned {len(results)} results.")

# Update the Search Object
new_so.alter(
    name='Updated Search Object',
    description='This is an updated description',
    name_query=NAME_QUERY_2,
    root_folder_query=FOLDER_QUERY_2,
)

# Delete the Search Object
new_so.delete(force=True)



"""This is the demo script to show how to manage prompts.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""

from pprint import pprint
from mstrio.connection import get_connection
from mstrio.modeling import Prompt, list_prompts, PromptRestrictions
from mstrio.project_objects import Report
from mstrio.types import ObjectSubTypes
from mstrio.modeling.schema.helpers import ObjectSubType


PROJECT_NAME = "MicroStrategy Tutorial"
conn = get_connection(workstationData, project_name=PROJECT_NAME)

# Try different filtering options in list_prompts function
all_prompts = list_prompts(conn)

name_filtered_prompts = list_prompts(conn, name='prompt')

prompts_dicts = list_prompts(conn, to_dictionary=True)

filtered_by_sub_type = list_prompts(conn, subtype=ObjectSubTypes.PROMPT_DOUBLE)

# Define values for new prompt
NEW_PROMPT_NAME = $new_prompt_name
NEW_PROMPT_DESTINATION_FOLDER = $new_prompt_destination_folder

# Create value prompt
simple_value_prompt = Prompt.create_value_prompt(
    conn,
    NEW_PROMPT_NAME,
    ObjectSubType.PROMPT_DOUBLE,
    destination_folder=NEW_PROMPT_DESTINATION_FOLDER,
    restrictions=PromptRestrictions(
        allow_personal_answers='single', required=True, min=1, max=100
    ),
)

# Init created prompt by name
existing_value_prompt = Prompt(conn, name=NEW_PROMPT_NAME)

# List properties
pprint(existing_value_prompt.list_properties())

# Alter existing prompt
existing_value_prompt.alter(name='Altered Prompt Name 3', default_answer={'value': 25})

# Add personal answer
existing_value_prompt.add_personal_answer(50, 'my_personal_answer')

# Delete existing prompt
existing_value_prompt.delete(force=True)

# Define values for new attribute elements prompt
NEW_ATTRIBUTE_ELEMENTS_PROMPT_NAME = $new_attribute_elements_prompt_name
SELECTED_ATTRIBUTE = $selected_attribute

# Create attribute elements prompt
attribute_elements_prompt = Prompt.create_attr_elements_prompt(
    conn,
    NEW_ATTRIBUTE_ELEMENTS_PROMPT_NAME,
    attribute=SELECTED_ATTRIBUTE,
    destination_folder=NEW_PROMPT_DESTINATION_FOLDER,
    restrictions=PromptRestrictions(allow_personal_answers='single', required=True),
    default_answer=[{'display': 'Electronics', 'elementId': 'h2'}],
)

attribute_elements_prompt.delete(force=True)



"""
    This is a template for creation of Datasource Script.

    Note:
        Also, if you use this template to create a datasource for Transaction Dashboard,
        which will be based on `transaction_edit_users.py` script template,
        those two will work out of the box end-to-end.

    It will work out of the box as is, when all variables are filled in.

    Three methods need to be created for the workflow to work:
        browse(): retrieves the catalog information
        preview(): gets partial data for preview, data refine and schema change
        publish(): gets the data published and stores the data into a cube
    All three methods have to return data in the format of Pandas DataFrame.
"""

"""
    This example will create two tables of data:
    - all_users: contains all users in the environment
    - admin_users: contains all users with names starting with "admin" or "Admin"

    The data is retrieved using the mstrio-py library, which is installed in the default runtime.
"""

import pandas as pd
from mstrio.connection import Connection
from mstrio.users_and_groups.user import User, list_users

# Those are three variables required for the script to work,
# none of which can be prompted: they need to be saved and will not be edited
# after saving the script
URL = $base_url
USER = $username
PASS = $password


conn = Connection(base_url=URL, username=USER, password=PASS)


# This is just a helper method which converts raw data gathered by this script
# into a Pandas DataFrame, which is required by the datasource script workflow
def users2dataframe(users: list[User]) -> pd.DataFrame:
    """Convert a list of User objects to a Pandas DataFrame."""
    data = [
        {
            "id": str(user.id),
            "name": str(user.name),
            "enabled": '1' if user.enabled else '0'
        } for user in users
    ]
    return pd.DataFrame(data)


def get_admin_users(max_count = None) -> pd.DataFrame:
    """Retrieve admin users from the environment."""
    users = (
        list_users(conn, name_begins="admin", limit=max_count)
        + list_users(conn, name_begins="Admin", limit=max_count)
    )
    return users2dataframe(users)


def get_all_users(max_count = None) -> pd.DataFrame:
    """Retrieve all users from the environment."""
    users = list_users(conn, limit=max_count)
    return users2dataframe(users)


# Everything above are only helper methods and actions to retrieve data.
# Below are methods required specifically by the datasource script workflow.
def browse():
    """Retrieve the catalog information.

    Returns:
        a python dictionary where:
        - keys are the names of the tables in final datasource
        - values are the Pandas DataFrame objects with data for each table.

    Note:
        The values should retrieve as little amount of data as possible,
        just enough to show the structure of the data, in order to work performantly.
    """
    return { 'all_users': get_all_users(1), 'admin_users': get_admin_users(1) }


def preview(table_name, row_count):
    """Get partial data for preview, data refine and schema change.

    Args:
        table_name: name of the selected table. This will be provided to this script
            by UI when DataSource Script will be used in for example Report or Transaction Dashboard.
        row_count: number of rows to be retrieved. This will be provided to this script
            by UI when DataSource Script will be used in for example Report or Transaction Dashboard.

    Returns:
        a Pandas DataFrame object with `row_count` amount of rows.
    """
    if table_name == 'admin_users':
        return get_admin_users(row_count)
    if table_name == 'all_users':
        return get_all_users(row_count)
    return None


def publish(table_name):
    """Get the data published and store the data into the cube.

    Args:
        table_name: name of the selected table. This will be provided to this script
            by UI when DataSource Script will be used in for example Report or Transaction Dashboard.

    Returns:
        a Pandas DataFrame object with all data.
    """
    if table_name == 'admin_users':
        return get_admin_users()
    if table_name == 'all_users':
        return get_all_users()
    return None



from typing import TYPE_CHECKING

from mstrio.utils.error_handlers import ErrorHandler

if TYPE_CHECKING:
    from mstrio.connection import Connection


@ErrorHandler(err_msg="Error creating a new changeset.")
def create_changeset(
    connection: 'Connection',
    project_id: str = None,
    schema_edit: bool = False,
    error_msg: str = None,
):
    """Create a new changeset for modelling manipulations."""
    return connection.post(
        endpoint='/api/model/changesets',
        headers={'X-MSTR-ProjectID': project_id},
        params={'schemaEdit': str(schema_edit).lower()},
    )


@ErrorHandler(err_msg="Error committing changeset {id} changes to the metadata.")
def commit_changeset_changes(
    connection: 'Connection', id: str, error_msg: str = None, throw_error=True
):
    """Commit the changeset changes to metadata."""
    return connection.post(
        endpoint=f'/api/model/changesets/{id}/commit',
        headers={'X-MSTR-MSChanget': id},
        params={'changesetId': id},
    )


@ErrorHandler(err_msg="Error deleting the changeset with ID {id}")
def delete_changeset(connection: 'Connection', id: str, error_msg: str = None):
    """Delete the changeset."""
    return connection.delete(
        endpoint=f'/api/model/changesets/{id}',
        headers={'X-MSTR-MSChanget': id},
        params={'changesetId': id},
    )



from mstrio.api import subscriptions as subscriptions_api
from mstrio.connection import Connection
from mstrio.helpers import IServerError
from mstrio.utils.version_helper import is_server_min_version


def get_subscription_status(connection: Connection, id: str):
    response = subscriptions_api.get_subscription_status(
        connection=connection, id=id, whitelist=[('ERR002', 500)]
    )
    res = response.json()
    server_msg = res.get('message')

    if not server_msg:
        return {'status': res}

    if (
        'No status for the subscription' in server_msg
        or 'This endpoint is disabled' in server_msg
    ):
        return {'status': None}

    server_code = res.get('code')
    ticket_id = res.get('ticketId')
    raise IServerError(
        message=f"{server_msg}; code: '{server_code}', ticket_id: '{ticket_id}'",
        http_code=response.status_code,
    )


def get_subscription_last_run(connection: Connection, id: str, project_id: str):
    if not is_server_min_version(connection, '11.4.0600'):
        return None

    response = subscriptions_api.list_subscriptions(
        connection=connection, project_id=project_id, last_run=True
    ).json()['subscriptions']
    sub = next(sub for sub in response if sub['id'] == id).get('lastRun')

    return {'last_run': sub}



from enum import Enum


# A sentinel object to detect if an attribute has been returned by REST API
class _MissingType:
    pass


MISSING = _MissingType()


class ObjectTypes(Enum):
    # https://www2.microstrategy.com/producthelp/Current/ReferenceFiles/reference/com/microstrategy/webapi/EnumDSSXMLObjectTypes.html
    FILTER = 1
    TEMPLATE = 2
    REPORT_DEFINITION = 3
    METRIC = 4
    AUTO_STYLE = 6
    AGG_METRIC = 7
    FOLDER = 8
    SUBSCRIPTION_DEVICE = 9
    PROMPT = 10
    FUNCTION = 11
    ATTRIBUTE = 12
    FACT = 13
    DIMENSION = 14
    TABLE = 15
    DATAMART_REPORT = 16
    FACT_GROUP = 17
    SHORTCUT_TYPE = 18
    RESOLUTION = 19
    MONITOR = 20
    ATTRIBUTE_FORM = 21
    SCHEMA = 22
    FORMAT = 23
    CATALOG = 24
    CATALOG_DEFINITION = 25
    COLUMN = 26
    PROPERTY_SET = 28
    DBROLE = 29
    DBLOGIN = 30
    DBCONNECTION = 31
    PROJECT = 32
    SERVER_DEFINITION = 33
    USER = 34
    USERGROUP = 34
    SUBSCRIPTION_TRANSMITTER = 35
    CONFIGURATION = 36
    REQUEST = 37
    SEARCH = 39
    SEARCH_FOLDER = 40
    DATAMART = 41
    PACKAGE_DEFINITION = 42
    ROLE = 43
    SECURITY_ROLE = 44
    LOCALE = 45
    CONSOLIDATION = 47
    CONSOLIDATION_ELEMENT = 48
    SCHEDULE_EVENT = 49
    SCHEDULE_OBJECT = 50
    SCHEDULE_TRIGGER = 51
    LINK = 52
    DBTABLE = 53
    TABLE_SOURCE = 54
    DOCUMENT_DEFINITION = 55
    DRILL_MAP = 56
    DBMS = 57
    SECURITY_FILTER = 58
    PROMPT_ANSWER = 59
    PROMPT_ANSWERS = 60
    GRAPH_STYLE = 61
    BLOB = 63
    OBJECT_TAG = 65
    CHANGE_JOURNAL = 66
    SHORTCUT = 67
    SHORTCUT_TARGET = 68
    RECONCILIATION = 69
    LAYER = 70
    PALETTE = 71
    THRESHOLDS = 72
    DASHBOARD_PERSONAL_VIEW = 73
    FEATURE_FLAG = 74
    SCRIPT = 76
    CONTENT_BUNDLE = 77
    APPLICATION = 78
    TIMEZONE = 79
    RUNTIME = 80
    CALENDAR = 81
    IAM = 82
    KPI_WATCHER = 83
    DRIVER = 84
    INTERFACE_LANGUAGE = 85
    WORKFLOW = 86
    TEST_SUITE = 93
    NOT_SUPPORTED = None

    def __new__(cls, value):
        member = object.__new__(cls)
        member._value_ = value
        return member

    def __int__(self):
        return self.value

    @classmethod
    def _rest_aliases(cls):
        return {
            ObjectTypes.AUTO_STYLE: 'style',
            ObjectTypes.SUBSCRIPTION_DEVICE: 'device',
            ObjectTypes.SHORTCUT_TYPE: 'type_shortcut',
            ObjectTypes.DBROLE: 'db_role',
            ObjectTypes.DBLOGIN: 'db_login',
            ObjectTypes.DBCONNECTION: 'db_connection',
            ObjectTypes.SERVER_DEFINITION: 'server_def',
            ObjectTypes.SUBSCRIPTION_TRANSMITTER: 'transmitter',
            ObjectTypes.PACKAGE_DEFINITION: 'function_package_definition',
            ObjectTypes.DBTABLE: 'db_table',
            ObjectTypes.DBMS: 'type_dbms',
            ObjectTypes.SECURITY_FILTER: 'md_security_filter',
            ObjectTypes.SHORTCUT: 'external_shortcut',
            ObjectTypes.SHORTCUT_TARGET: 'external_shortcut_target',
            ObjectTypes.DASHBOARD_PERSONAL_VIEW: 'dossier_personal_view',
            ObjectTypes.SCRIPT: 'command_manager_script',
            ObjectTypes.RUNTIME: 'script_runtime_env',
        }

    @classmethod
    def _missing_(cls, value):
        aliases = cls._rest_aliases()
        aliases_to_enum_items = {aliases[enum_item]: enum_item for enum_item in aliases}
        if value in aliases_to_enum_items:
            return aliases_to_enum_items[value]
        else:
            member = object.__new__(cls)
            member._value_ = value
            member._name_ = f'UNDEFINED_TYPE:{value}'
            return member

    @staticmethod
    def contains(item):
        if isinstance(item, ObjectTypes):
            return True
        return item in [v.value for v in ObjectTypes.__members__.values()]

    def get_subtypes(self):
        # This takes advantage of how subtypes are defined in DSS.
        # All subtypes are defined as the corresponding type followed
        # by 8 bits of subtype ID
        return [
            subtype
            for subtype in ObjectSubTypes
            if subtype.value and subtype.value >> 8 == self.value
        ]

    @classmethod
    def from_rest_value(cls, source, *args, **kwargs):
        """Handle missing values in the enum."""
        aliases = cls._rest_aliases()
        aliases_to_enum_items = {aliases[enum_item]: enum_item for enum_item in aliases}
        if source in aliases_to_enum_items:
            return aliases_to_enum_items[source]
        else:
            return cls[source.upper()]

    def to_rest_value(self):
        """Return the REST alias for the object subtype."""
        aliases = self._rest_aliases()
        return aliases.get(self, self.name.lower())


class ObjectSubTypes(Enum):
    # https://www2.microstrategy.com/producthelp/Current/ReferenceFiles/reference/com/microstrategy/webapi/EnumDSSXMLObjectSubTypes.html
    FILTER = 256
    CUSTOM_GROUP = 257
    TEMPLATE = 512
    REPORT_GRID = 768
    REPORT_GRAPH = 769
    REPORT_ENGINE = 770
    REPORT_TEXT = 771
    REPORT_DATAMART = 772
    REPORT_BASE = 773
    REPORT_GRID_AND_GRAPH = 774
    REPORT_NON_INTERACTIVE = 775
    OLAP_CUBE = 776
    INCREMENTAL_REFRESH_REPORT = 777
    REPORT_TRANSACTION = 778
    SUPER_CUBE = 779
    SUPER_CUBE_IRR = 780
    REPORT_HYPER_CARD = 781
    METRIC = 1024
    SUBTOTAL_DEFINITION = 1025
    SYSTEM_SUBTOTAL = 1026
    METRIC_DMX = 1027
    METRIC_TRAINING = 1028
    STYLE = 1536
    METRIC_AGG = 1792
    FOLDER = 2048
    DEVICE = 2304
    PROMPT = 2560
    PROMPT_BOOLEAN = 2561
    PROMPT_LONG = 2562
    PROMPT_STRING = 2563
    PROMPT_DOUBLE = 2564
    PROMPT_DATE = 2565
    PROMPT_OBJECTS = 2566
    PROMPT_ELEMENTS = 2567
    PROMPT_EXPRESSION = 2568
    PROMPT_EXPRESSION_DRAFT = 2569
    PROMPT_DIMTY = 2570
    PROMPT_BIG_DECIMAL = 2571
    FUNCTION = 2816
    ATTRIBUTE = 3072
    ATTRIBUTE_ROLE = 3073
    ATTRIBUTE_TRANSFORMATION = 3074
    ATTRIBUTE_ABSTRACT = 3075
    ATTRIBUTE_RECURSIVE = 3076
    ATTRIBUTE_DERIVED = 3077
    ATTRIBUTE_SMART = 3078
    FACT = 3328
    DIMENSION_SYSTEM = 3584
    DIMENSION_USER = 3585
    DIMENSION_ORDERED = 3586
    DIMENSION_USER_HIERARCHY = 3587
    TABLE = 3840
    TABLE_PARTITION_MD = 3841
    TABLE_PARTITION_WH = 3842
    DATAMART_REPORT = 4096
    FACT_GROUP = 4352
    SHORTCUT = 4608
    LIBRARY_SHORTCUT = 4609
    RESOLUTION = 4864
    ATTRIBUTE_FORM = 5376
    ATTRIBUTE_FORM_SYSTEM = 5377
    ATTRIBUTE_FORM_NORMAL = 5378
    SCHEMA = 5632
    FORMAT = 5888
    CATALOG = 6144
    CATALOG_DEFINITION = 6400
    COLUMN = 6656
    PROPERTY_GROUP = 6912
    PROPERTY_SET = 7168
    DB_ROLE = 7424
    DB_ROLE_DATA_IMPORT = 7425
    DB_ROLE_DATA_IMPORT_PRIMARY = 7426
    DB_ROLE_OAUTH = 7427
    DB_ROLE_REMOTE_DATA_SOURCE = 7428
    DB_ROLE_URL_AUTH = 7429
    DB_ROLE_GENERIC_DATA_CONNECTOR = 7430
    DB_LOGIN = 7680
    DB_LOGIN_SECRET_VAULT = 7681
    DB_CONNECTION = 7936
    PROJECT = 8192
    SERVER_DEFINITION = 8448
    USER = 8704
    USER_GROUP = 8705
    TRANSMITTER = 8705
    CONFIGURATION = 9216
    REQUEST = 9472
    SEARCH = 9984
    SEARCH_FOLDER = 10240
    DATAMART = 10496
    FUNCTION_PACKAGE_DEFINITION = 10752
    ROLE = 11008
    ROLE_TRANSFORMATION = 11009
    SECURITY_ROLE = 11264
    INBOX = 11520
    INBOX_MESSAGE = 11776
    CONSOLIDATION = 12032
    CONSOLIDATION_MANAGED = 12033
    CONSOLIDATION_ELEMENT = 12288
    SCHEDULE_EVENT = 12544
    SCHEDULE_OBJECT = 12800
    SCHEDULE_TRIGGER = 13056
    LINK = 13312
    DB_TABLE = 13568
    DB_TABLE_PMT = 13569
    TABLE_SOURCE = 13824
    DOCUMENT_DEFINITION = 14080
    REPORT_WRITING_DOCUMENT = 14081
    DOCUMENT_THEME = 14082
    DOCUMENT_BOT = 14084
    DOCUMENT_BOT_2_0 = 14087
    DOCUMENT_BOT_UNIVERSAL = 14091
    DOCUMENT_AGENT = 14087
    DOCUMENT_AGENT_UNIVERSAL = 14091
    DRILL_MAP = 14336
    DBMS = 14592
    MD_SECURITY_FILTER = 14848
    MONITOR_PERFORMANCE = 15104
    MONITOR_JOBS = 15105
    MONITOR_USER_CONNECTIONS = 15106
    MONITOR_DB_CONNECTIONS = 15107
    PROMPT_ANSWER = 15232
    PROMPT_ANSWERS = 15360
    PROMPT_ANSWER_BOOLEAN = 15233
    PROMPT_ANSWER_LONG = 15234
    PROMPT_ANSWER_STRING = 15235
    PROMPT_ANSWER_DOUBLE = 15236
    PROMPT_ANSWER_DATE = 15237
    PROMPT_ANSWER_OBJECTS = 15238
    PROMPT_ANSWER_ELEMENTS = 15239
    PROMPT_ANSWER_EXPRESSION = 15240
    PROMPT_ANSWER_EXPRESSION_DRAFT = 15241
    PROMPT_ANSWER_DIMTY = 15242
    PROMPT_ANSWER_BIG_DECIMAL = 15243
    PROMPT_ANSWER_INT64 = 15244
    GRAPH_STYLE = 15616
    CHANGE_JOURNAL_SEARCH = 15872
    BLOB_UNKNOWN = 16128
    BLOB_OTHER = 16129
    BLOB_IMAGE = 16130
    BLOB_PROJECT_PACKAGE = 16131
    BLOB_EXCEL = 16132
    BLOB_HTML_TEMPLATE = 16133
    DASHBOARD_TEMPLATE = 16384
    FLAG = 16640
    CHANGE_JOURNAL = 16896
    EXTERNAL_SHORTCUT_UNKNOWN = 17152
    EXTERNAL_SHORTCUT_URL = 17153
    EXTERNAL_SHORTCUT_SNAPSHOT = 17154
    EXTERNAL_SHORTCUT_TARGET = 17408
    RECONCILIATION = 17664
    SYSTEM_PALETTE = 17920
    CUSTOM_PALETTE = 17921
    THRESHOLDS = 18432
    COMMAND_MANAGER_SCRIPT = 19456
    SCRIPT = 19457
    JUPYTER_NOTEBOOK_SCRIPT = 19458
    DATASOURCE_SCRIPT = 19459
    TRANSACTION_SCRIPT = 19460
    APPLICATION = 19968
    APPLICATION_POWERPOINT = 19969
    APPLICATION_EXCEL = 19970
    APPLICATION_TEAMS = 19971
    APPLICATION_TABLEAU = 19972
    APPLICATION_GOOGLE_SHEETS = 19973
    TIMEZONE_SYSTEM = 20224
    TIMEZONE_CUSTOM = 20225
    CALENDAR_SYSTEM = 20736
    CALENDAR_CUSTOM = 20737
    INTERFACE_LANGUAGE_SYSTEM = 21760
    INTERFACE_LANGUAGE_CUSTOM = 21761
    WORKFLOW = 22016
    IAM_SECRET_VAULT = 21002
    NUGGETS_FILE = 23040
    STREAM_CHANNEL = 23552
    SUBSCRIPTION_ADDRESS = 65281
    SUBSCRIPTION_CONTACT = 65282
    SUBSCRIPTION_INSTANCE = 65283
    NOT_SUPPORTED = None

    def __new__(cls, value):
        member = object.__new__(cls)
        member._value_ = value
        return member

    def __int__(self):
        return self.value

    @classmethod
    def _rest_aliases(cls):
        return {
            ObjectSubTypes.OLAP_CUBE: 'report_cube',
            ObjectSubTypes.INCREMENTAL_REFRESH_REPORT: 'report_incremental_refresh',
            ObjectSubTypes.SUPER_CUBE: 'report_emma_cube',
            ObjectSubTypes.SUPER_CUBE_IRR: 'report_emma_incremental_refresh',
            ObjectSubTypes.REPORT_HYPER_CARD: 'report_hypercard',
            ObjectSubTypes.METRIC_AGG: 'agg_metric',
            ObjectSubTypes.PROMPT_BIG_DECIMAL: 'prompt_bigdecimal',
            ObjectSubTypes.ATTRIBUTE_DERIVED: 'derived_attribute',
            ObjectSubTypes.ATTRIBUTE_FORM_SYSTEM: 'form_system',
            ObjectSubTypes.ATTRIBUTE_FORM_NORMAL: 'form_normal',
            ObjectSubTypes.CATALOG_DEFINITION: 'catalog_def',
            ObjectSubTypes.DB_ROLE_GENERIC_DATA_CONNECTOR: (
                'db_role_generic_dataconnector'
            ),
            ObjectSubTypes.SERVER_DEFINITION: 'server_def',
            ObjectSubTypes.FUNCTION_PACKAGE_DEFINITION: 'function_package_def',
            ObjectSubTypes.INBOX_MESSAGE: 'inbox_msg',
            ObjectSubTypes.DOCUMENT_DEFINITION: 'document_def',
            ObjectSubTypes.DOCUMENT_BOT: 'chat_bot',
            ObjectSubTypes.DOCUMENT_BOT_2_0: 'chat_bot_v2',
            ObjectSubTypes.DOCUMENT_BOT_UNIVERSAL: 'universal_chat_bot',
            ObjectSubTypes.DOCUMENT_AGENT: 'chat_bot_v2',
            ObjectSubTypes.DOCUMENT_AGENT_UNIVERSAL: 'universal_chat_bot',
            ObjectSubTypes.PROMPT_ANSWER_BIG_DECIMAL: 'prompt_answer_bigdecimal',
            ObjectSubTypes.CHANGE_JOURNAL_SEARCH: 'changejournal_search',
            ObjectSubTypes.CHANGE_JOURNAL: 'changejournal',
            ObjectSubTypes.SYSTEM_PALETTE: 'palette_system',
            ObjectSubTypes.CUSTOM_PALETTE: 'palette_custom',
            ObjectSubTypes.SCRIPT: 'command_manager_script_python',
            ObjectSubTypes.JUPYTER_NOTEBOOK_SCRIPT: (
                'command_manager_script_jupyter_notebook'
            ),
            ObjectSubTypes.IAM_SECRET_VAULT: 'vault_connection',
        }

    @classmethod
    def from_rest_value(cls, source, *args, **kwargs):
        """Handle missing values in the enum."""
        aliases = cls._rest_aliases()
        aliases_to_enum_items = {aliases[enum_item]: enum_item for enum_item in aliases}
        if source in aliases_to_enum_items:
            return aliases_to_enum_items[source]
        else:
            return cls[source.upper()]

    def to_rest_value(self):
        """Return the REST alias for the object subtype."""
        aliases = self._rest_aliases()
        return aliases.get(self, self.name.lower())


class ExtendedType(Enum):
    RESERVED = 0
    RELATIONAL = 1
    MDX = 2
    CUSTOM_SQL_FREE_FORM = 3
    CUSTOM_SQL_WIZARD = 4
    FLAT_FILE = 5
    DATA_IMPORT_GENERAL = 0x100
    DATA_IMPORT_FILE_EXCEL = 0x110
    DATA_IMPORT_FILE_TEXT = 0x120
    DATA_IMPORT_CUSTOM_SQL = 0x130
    DATA_IMPORT_TABLE = 0x140
    DATA_IMPORT_OAUTH = 0x150
    DATA_IMPORT_OAUTH_SFDC = 0x151
    DATA_IMPORT_OAUTH_GDRIVE = 0x152
    DATA_IMPORT_OAUTH_DROPBOX = 0x153
    _ = 0x155
    DATA_IMPORT_CUSTOM_SQL_WIZARD = 0x160
    DATA_IMPORT_SINGLE_TABLE = 0x170
    DATA_IMPORT_SPARK = 0x1A0
    DATA_IMPORT_DATASET = 0x1C0
    DATA_IMPORT_OLAP_CUBE = 0x1000
    DATA_IMPORT_LIVE_CUBE = 0x1001

    def __new__(cls, value):
        member = object.__new__(cls)
        member._value_ = value
        return member

    def __int__(self):
        return self.value


TypeOrSubtype = (
    int | ObjectTypes | ObjectSubTypes | list[int | ObjectTypes | ObjectSubTypes]
)



"""
Sample script showing how to duplicate a project.

This script uses MicroStrategy Tutorial project. Please replace its values
to use it with your own project.

1. Connect to the environment using data from the workstation
2. Define the name of the project to duplicate
3. Select project languages for duplication
4. Define non default duplication settings
5. Perform project duplication
6. Wait for the duplication to finish
7. Print the status of the duplication

"""

from mstrio.connection import get_connection
from mstrio.server import DuplicationConfig, Project, list_languages

conn = get_connection(workstationData)


PROJECT_NAME = 'MicroStrategy Tutorial' # Insert the name of project to duplicate
base_project = Project(connection=conn, name=PROJECT_NAME)

# Insert name of duplicated project
DUPLICATION_NAME = 'MicroStrategy Tutorial duplication 1'

# Duplicate the project with no default settings
# List languages to select from in duplication settings
langs = list_languages(conn)

selected_langs = langs[:5]  # Select first 5 languages for duplication
selected_default_project_lang= langs[0].lcid  # Select first language as default locale

# Create a list of import locales consisting from lcids of selected languages

selected_langs = [lang.lcid for lang in selected_langs]

# Define not default duplication settings
duplication_config= DuplicationConfig(schema_objects_only=False,
                                      skip_empty_profile_folders=False,
                                      include_user_subscriptions=False,
                                      include_contact_subscriptions=False,
                                      import_description="test duplication with languages",
                                      import_default_locale=selected_default_project_lang,
                                      import_locales=selected_langs)

# Perform duplication
# Target name should be unique, otherwise request will fail
duplication = base_project.duplicate(
    target_name=DUPLICATION_NAME, duplication_config=duplication_config
)

duplication.wait_for_stable_status()

print(
    f"Project duplication '{DUPLICATION_NAME}' (ID: {duplication.id}) has finished with status: {duplication.status.value}"
)



"""
    This is a template for creation of Transaction Script.

    Specifically, if the Transaction Dashboard is based on Python Script Datasource
    built from `data_source_script_all_users.py` template, those two will work
    out of the box end-to-end.

    It will work out of the box as is, as long as the Transaction Dashboard
    is configured correctly and all variables are set up.
"""

"""
    This example will allow a dashboard user to modify users in the environment.
    The workflow will rely on this script to make actual modifications
    (insert, update or delete) in the users.

    The data is modified using the mstrio-py library, which is installed in the default runtime.

    No additional external libraries are required for this script to work.
"""

import sys
from mstrio.connection import Connection
from mstrio.users_and_groups.user import User

# Those are three variables required for the script to work,
# none of which can be prompted: they need to be saved and will not be edited
# after saving the script
URL = $base_url
USER = $username
PASS = $password


conn = Connection(base_url=URL, username=USER, password=PASS)


# It is a good practice to `zip` all transaction columns like this, so that
# the script can handle them in a single loop, instead of having to
# handle each column separately.
#
# But this is not a requirement.
#
# Also, this set of variables are only an example. Transaction Scripts can decide
# any number and any type of columns to handle, as long as they are set up
# correctly in the Transaction Dashboard.
#
# In this example, we assume that the Transaction Dashboard consumer will pass
# the following columns:
# - txn_action: a list of actions to perform (insert, update, or select)
#       (this is a built-in type of column and is always provided from Transaction Dashboard)
# - user_id: a list of user IDs to read or modify
# - user_name: a list of user names to read or modify
# - user_enabled: a list of user enabled status ('1' for enabled, '0' for disabled)
rows = zip($txn_action, $user_id, $user_name, $user_enabled)


for action, uid, uname, uenabled in rows:
    enabled = uenabled == '1'  # we're converting a string '1' or '0' to a boolean value

    try:
        # `try` here is a good practice - we should always assume something can go wrong,
        # but the script should not crash, so we need to handle it

        if action == 'select':
            # if Transaction Dashboard consumer just selected a row in Dashboard without changes in it:
            # it should be deleted
            user = User(conn, id=uid)
            if user:
                user.delete(force=True)

        elif action == 'insert':
            # Transaction Dashboard consumer wants to create a new user.

            # for this example specifically we build username from full name so we
            # remove spaces from it
            username = uname.replace(' ', '_')
            User.create(conn,
                username=username, full_name=uname,
                enabled=enabled,
                password='qwer1234!', require_new_password=True)

        elif action == 'update':
            # Transaction Dashboard consumer wants to alter the properties of existing user
            user = User(conn, id=uid)
            if user:
                user.alter(full_name=uname, enabled=enabled)

        else:
            # unexpected and unreachable - ignore and move to next row of data
            continue
    except Exception as err:
        # log the output of the error, if any happened, and proceed to next row of data
        print(f"Failed '{action}' for user '{uname}' (id={uid}): {str(err)}", file=sys.stderr)
        continue



# flake8: noqa
# isort: off
from .dimensionality import Dimensionality, DimensionalityUnit
from .metric_format import FormatProperty, MetricFormat
from .metric import (
    DefaultSubtotals,
    FormatProperty,
    list_metrics,
    Metric,
    MetricFormat,
    Threshold,
)

# isort: on



"""This is the demo script to show how to include custom logging messages
into your script, independently from what the code itself will log.
"""

import logging
import sys

from mstrio.connection import get_connection
from mstrio.datasources import list_datasource_instances

# At any point you can you use built-in `print` method to print something
# to console and, in turn, to the log file.
print('I am running MyCustomScript!')

# However, if you wish to customize what and how will be logged, you can use a
# built-in `logging` module and customize it for your script specifically.
LOGGER_REFERENCE = 'MyCustomScript'
logger = logging.getLogger(LOGGER_REFERENCE)

# You can customize what level of logging will be stored. Every level below it
# will just be ignored. Available levels (in order) are:
# DEBUG, INFO, WARNING, ERROR, CRITICAL
logger.setLevel(logging.INFO)  # we set the level to INFO, so DEBUG will be ignored

# The below set if instructions basically tell the script to do the following:
# 1. Whenever `logger` is used, print the message to console and keep it in logs
# 2. Format messages in a custom way including logger name and message's level
stream_handler = logging.StreamHandler(sys.stdout)
# (feel free to customize the formatting to your liking)
# (details in documentation: <https://docs.python.org/3/library/logging.html>)
stream_handler.setFormatter(logging.Formatter('[%(name)s]{%(levelname)s}: %(message)s'))
# (this level should always be DEBUG as this applies to message stream
# configuration and not the logger itself)
stream_handler.setLevel(logging.DEBUG)
logger.addHandler(stream_handler)

# Now you can use the logger to log messages of different levels.
#
# If you did not change anything above you'll notice that only messages INFO
# and above will be logged to console, while DEBUG message will be ignored.
# This is because of `logger.setLevel(logging.INFO)` line above.
logger.debug('This is DEBUG')
logger.info('This is INFO')
logger.warning('This is WARN')
logger.error('This is ERROR')
logger.critical('This is CRIT')


"""
Below is an example of logging use in actual workflow.

We're updating VLDB SQL username settings to '/*!u*/'
for all relevant datasource instances.
"""

vldb_names = [  # example names
    'Post Insert String',
    'SELECT PostString for Main SQL',
    'SELECT PostString',
]

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

formatter = logging.Formatter(
    '%(asctime)s - %(name)s - %(levelname)s - %(message)s',
    datefmt='%Y-%m-%d %H:%M:%S',
)
console_handler = logging.StreamHandler(sys.stdout)
console_handler.setLevel(logging.DEBUG)
console_handler.setFormatter(formatter)
logger.addHandler(console_handler)

conn = get_connection(workstationData)

try:
    datasource_instances = list_datasource_instances(connection=conn)
except Exception as e:
    logger.error(f"Failed to list datasource instances: {e}")
    raise e  # break the script by re-raising the error

for datasource in datasource_instances:
    try:
        vldb_settings = datasource.list_vldb_settings(
            to_dictionary=True,
            groups=['Select/Insert'],
        )
        relevant_vldbs = [vldb for vldb in vldb_settings if vldb in vldb_names]

        vldbs_to_update = {}
        for vldb in relevant_vldbs:
            current_value = datasource.vldb_settings[vldb].value
            logger.debug(f'{vldb}: {current_value}')
            if current_value != '/*!u*/':
                logger.debug(
                    f'Adding {vldb} for datasource {datasource.name} to update list'
                )
                vldbs_to_update[vldb] = '/*!u*/'

        if vldbs_to_update:
            logger.info(f"Updating '{datasource.name}' VLDBs: {list(vldb)} to '/*!u*/'")
            try:
                datasource.alter_vldb_settings(names_to_values=vldbs_to_update)
            except Exception as e:
                logger.error(
                    "Failed to update VLDB settings "
                    f"for datasource '{datasource.name}': {e}"
                )
                continue
    except Exception as e:
        logger.error(f"Failed to process datasource '{datasource.name}': {e}")



"""This is the demo script to show how to manage user hierarchies.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""

from mstrio.modeling.schema import (
    ElementDisplayOption,
    HierarchyAttribute,
    HierarchyRelationship,
    SchemaManagement,
    SchemaObjectReference,
    SchemaUpdateType,
    ObjectSubType,
    UserHierarchy,
    UserHierarchySubType,
    list_user_hierarchies
)

from mstrio.connection import get_connection

# For every object we want to reference using a SchemaObjectReference we need
# to provide an Object ID for. For the script to work correctly all occurrences
# of `'<object_id>'` and others with form `<some_name>` need to be replaced with
# data specific to the object used.

# Define a variable which can be later used in a script
PROJECT_NAME = $project_name  # Project to connect to

conn = get_connection(workstationData, project_name=PROJECT_NAME)

# get a list of user hierarchies
user_hierarchies = list_user_hierarchies(conn)

# Define variables which can be later used in a script
ATTRIBUTE_NAME_1 = $attribute_name_1
ATTRIBUTE_NAME_2 = $attribute_name_2
ATTRIBUTE_NAME_3 = $attribute_name_3
ATTRIBUTE_NAME_4 = $attribute_name_4

# create attributes to be added to a user hierarchy
# an attribute can be either an object (HierarchyAttribute) or a dict
attribute1 = HierarchyAttribute(
    object_id='<object_id>',
    entry_point=True,
    name=ATTRIBUTE_NAME_1,
    # see modeling/schema/user_hierarchy/user_hierarchy.py for available options
    element_display_option=ElementDisplayOption.ALL_ELEMENTS,
)

# Define a variable which can be later used in a script
# see modeling/schema/user_hierarchy/user_hierarchy.py -
# - ElementDisplayOption class for available options
ELEMENT_DISPLAY_OPTION = $element_display_option

attribute2 = {
    "objectId": '<object_id>',
    "entryPoint": True,
    "name": ATTRIBUTE_NAME_2,
    "elementDisplayOption": ELEMENT_DISPLAY_OPTION
}

# create SchemaObjectReference objects to be used
# for relationship creation
sor1 = SchemaObjectReference(
    object_id='<object_id>',
    sub_type=ObjectSubType.ATTRIBUTE  # see modeling/schema/helpers.py - ObjectSubType class
    # for available options
)

sor2 = SchemaObjectReference(
    object_id='<object_id>',
    sub_type=ObjectSubType.ATTRIBUTE  # see modeling/schema/helpers.py - ObjectSubType class
    # for available options
)

# create a hierarchy relationship between two attributes
relationship = HierarchyRelationship(parent=sor1, child=sor2)

# Define variables which can be later used in a script
USER_HIERARCHY_NAME = $user_hierarchy_name
FOLDER_ID = $folder_id

# create a user hierarchy with subtype `DIMENSION_USER_HIERARCHY`
# and previously created attributes and relationship
new_user_hierarchy = UserHierarchy.create(
    connection=conn,
    name=USER_HIERARCHY_NAME,
    # see modeling/schema/user_hierarchy/user_hierarchy.py for available options
    sub_type=UserHierarchySubType.DIMENSION_USER_HIERARCHY,
    attributes=[attribute1, attribute2],
    relationships=[relationship],
    destination_folder_id=FOLDER_ID,
)

# Any changes to a schema objects must be followed by schema_reload
# in order to use them in reports, dashboards and so on
schema_manager = SchemaManagement(connection=conn, project_id=conn.project_id)
task = schema_manager.reload(update_types=[SchemaUpdateType.LOGICAL_SIZE])

# Define a variable which can be later used in a script
USER_HIERARCHY_DESCRIPTION = $user_hierarchy_description

# alter a user hierarchy, change it's name, description
# and subtype
new_user_hierarchy.alter(
    name=USER_HIERARCHY_NAME,
    description=USER_HIERARCHY_DESCRIPTION,
    # see modeling/schema/user_hierarchy/user_hierarchy.py for available options
    sub_type=UserHierarchySubType.DIMENSION_USER,
    use_as_drill_hierarchy=False
)

# define additional attributes
attribute3 = HierarchyAttribute(
    object_id='<object_id>',
    entry_point=True,
    name=ATTRIBUTE_NAME_3,
    element_display_option=ElementDisplayOption.ALL_ELEMENTS,
)

attribute4 = HierarchyAttribute(
    object_id='<object_id>',
    entry_point=False,
    name=ATTRIBUTE_NAME_4,
    element_display_option=ElementDisplayOption.ALL_ELEMENTS,
    limit=50
)

# add new attributes to a hierarchy
new_user_hierarchy.add_attribute(attribute3)
new_user_hierarchy.add_attribute(attribute4)

# define another relationship (it may be a dict)
relationship = {
    'parent': {
        'objectId': '<object_id>',
        'subType': ObjectSubType.ATTRIBUTE  # see modeling/schema/helpers.py for available values
    },
    'child': {
        'objectId': '<object_id>',
        'subType': ObjectSubType.ATTRIBUTE  # see modeling/schema/helpers.py for available values
    }
}

# add new relationship to a hierarchy
new_user_hierarchy.add_relationship(relationship)

# remove a relationship from a hierarchy
new_user_hierarchy.remove_relationship(relationship)

# remove an attribute from a hierarchy
new_user_hierarchy.remove_attribute(attribute3)

# get a user hierarchy by ID. User hierarchy can be also found by its name.
user_hierarchy = UserHierarchy(conn, id=new_user_hierarchy.id)
user_hierarchy_by_name = UserHierarchy(conn, name=USER_HIERARCHY_NAME)

# list properties for user hierarchy
user_hierarchy.list_properties()

# Delete a user_hierarchy.
# When argument `force` is set to `False` (default value),
# then deletion must be confirmed by selecting appropriate prompt value.
user_hierarchy.delete(force=True)



from mstrio.api import usergroups as usergroups_api
from mstrio.utils.helper import camel_to_snake
from mstrio.utils.version_helper import is_server_min_version


def update_user_group_info(connection, id, body):
    """Update user group by a specified ID.

    Args:
        connection: Strategy One REST API connection object
        id: ID of the user group
        body: dict with changes to be made

    Returns:
        dict with user group information
    """
    # We do not want information about members from patch as it is insufficient
    # to differentiate between users and user groups
    response = usergroups_api.update_user_group_info(
        connection=connection, id=id, body=body
    )
    result = response.json()
    if result.get('members'):
        result.pop('members')

    return result


def get_settings(
    connection, id, include_access=False, offset=0, limit=-1, error_msg=None
):
    response = usergroups_api.get_settings(
        connection=connection,
        id=id,
        include_access=include_access,
        offset=offset,
        limit=limit,
        error_msg=error_msg,
    ).json()

    if is_server_min_version(connection, '11.4.1200'):
        list_of_settings = [
            {key: value} for key, value in camel_to_snake(response).items()
        ]
        response = {'settings': list_of_settings}

    return response



import contextlib
from typing import TYPE_CHECKING

from mstrio.api import library
from mstrio.connection import Connection
from mstrio.helpers import VersionException
from mstrio.object_management.library_shortcut import LibraryShortcut
from mstrio.project_objects.agents import Agent, list_agents
from mstrio.project_objects.dashboard import Dashboard, list_dashboards
from mstrio.project_objects.document import Document, list_documents
from mstrio.project_objects.report import Report, list_reports
from mstrio.utils.helper import deprecation_warning
from mstrio.utils.resolvers import get_project_id_from_params_set

if TYPE_CHECKING:
    from mstrio.server.project import Project


class Library:
    """Representation of the state of the authenticated user's Library.
    It contains the references to the documents, dashboards, and reports
    that have been shared to the user's Library.

    Attributes:
        connection (Connection): Strategy One connection object returned
            by `connection.Connection()`.
        project_id (str): ID of the project that the Library is in.
        user_id (str): ID of the authenticated user.
        documents (list[Document]): List of documents in the Library.
        dashboards (list[Dashboard]): List of dashboards in the Library.
        contents (list[Document | Dashboard]): List of all contents
            (documents and dashboards) in the Library.
    """

    def __init__(
        self,
        connection: Connection,
        project: 'Project | str | None' = None,
        project_id: str | None = None,
        project_name: str | None = None,
    ):
        """Initialize the Library object.

        Args:
            connection (Connection): Strategy One connection object returned
                by `connection.Connection()`.
            project (Project | str, optional): Project object or ID or name
                specifying the project. May be used instead of `project_id` or
                `project_name`.
            project_id (str, optional): Project ID
            project_name (str, optional): Project name
        """
        self.connection = connection
        self.user_id = connection.user_id

        self.project_id = get_project_id_from_params_set(
            connection,
            project,
            project_id,
            project_name,
        )

        # Shortcut data dicts, as they are retrieved from the Library API.
        # They are parsed directly into LibraryShortcut objects. Documents and
        # Dashboards are grouped together in the response, hence the joint dict
        self._doc_dbd_shortcuts_data: list[dict] = []
        self._report_shortcuts_data: list[dict] = []
        self._agent_shortcuts_data: list[dict] = []

        # Cached shortcut IDs for quick to avoid unnecessary retrieval of target
        # objects after the shortcuts list changes
        self._doc_dbd_cached_short_ids: set[str] = set()
        self._report_cached_short_ids: set[str] = set()
        self._agent_cached_short_ids: set[str] = set()

        # The shortcuts themselves and corresponding target objects
        self._documents: list[Document] = []
        self._dashboards: list[Dashboard] = []
        self._reports: list[Report] = []
        self._agents: list[Agent] = []
        self._document_shortcuts: list[LibraryShortcut] = []
        self._dashboard_shortcuts: list[LibraryShortcut] = []
        self._report_shortcuts: list[LibraryShortcut] = []
        self._agent_shortcuts: list[LibraryShortcut] = []

        self.fetch()

    def fetch(self):
        response: dict = library.get_library_v2(self.connection).json()
        self._doc_dbd_shortcuts_data = response.get('documentContents', [])
        self._report_shortcuts_data = response.get('reportContents', [])
        self._agent_shortcuts_data = response.get('aiBotContents', [])

    @staticmethod
    def _get_target_ids(shortcuts_data: list[dict]) -> list[str]:
        """Extracts target IDs from the shortcuts data."""
        return [
            sho.get('target', {}).get('id') for sho in shortcuts_data if 'target' in sho
        ]

    def _set_documents_dashboards(self, short_ids: set[str]):
        doc_ids = self._get_target_ids(self._doc_dbd_shortcuts_data)
        self._documents = list_documents(
            self.connection,
            project_id=self.project_id,
            id=doc_ids,
        )
        self._dashboards = list_dashboards(
            self.connection,
            project_id=self.project_id,
            id=doc_ids,
        )
        self._doc_dbd_cached_short_ids = short_ids

        document_short_ids = [doc.id for doc in self._documents]
        self._document_shortcuts = [
            LibraryShortcut.from_dict(sho, self.connection)
            for sho in self._doc_dbd_shortcuts_data
            if sho.get('target', {}).get('id') in document_short_ids
        ]
        dashboard_short_ids = [dbd.id for dbd in self._dashboards]
        self._dashboard_shortcuts = [
            LibraryShortcut.from_dict(sho, self.connection)
            for sho in self._doc_dbd_shortcuts_data
            if sho.get('target', {}).get('id') in dashboard_short_ids
        ]

    @property
    def documents(self):
        doc_dbd_live_short_ids = {doc['id'] for doc in self._doc_dbd_shortcuts_data}
        if doc_dbd_live_short_ids != self._doc_dbd_cached_short_ids:
            self._set_documents_dashboards(doc_dbd_live_short_ids)
        return self._documents

    @property
    def dashboards(self):
        doc_dbd_live_short_ids = {doc['id'] for doc in self._doc_dbd_shortcuts_data}
        if doc_dbd_live_short_ids != self._doc_dbd_cached_short_ids:
            self._set_documents_dashboards(doc_dbd_live_short_ids)
        return self._dashboards

    @property
    def reports(self):
        report_live_short_ids = {rep['id'] for rep in self._report_shortcuts_data}
        if report_live_short_ids != self._report_cached_short_ids:
            report_ids = self._get_target_ids(self._report_shortcuts_data)
            self._reports = list_reports(
                self.connection,
                project_id=self.project_id,
                id=report_ids,
            )
            self._report_cached_short_ids = report_live_short_ids
            self._report_shortcuts = [
                LibraryShortcut.from_dict(sho, self.connection)
                for sho in self._report_shortcuts_data
            ]
        return self._reports

    @property
    def bots(self):
        deprecation_warning(
            'Library.bots', 'Library.agents', '11.6.1.101', module=False  # NOSONAR
        )
        return self.agents

    @property
    def agents(self):
        agent_live_short_ids = {agent['id'] for agent in self._agent_shortcuts_data}
        if agent_live_short_ids != self._agent_cached_short_ids:
            agent_ids = self._get_target_ids(self._agent_shortcuts_data)
            with contextlib.suppress(VersionException):
                # Leave the agents field empty if no support
                self._agents = list_agents(
                    self.connection,
                    project_id=self.project_id,
                    id=agent_ids,
                )
            self._agent_cached_short_ids = agent_live_short_ids
            self._agent_shortcuts = [
                LibraryShortcut.from_dict(sho, self.connection)
                for sho in self._agent_shortcuts_data
            ]
        return self._agents

    @property
    def contents(self):
        return self.documents + self.dashboards + self.reports + self.agents

    def publish(self, contents: "list | Dashboard | Document | str"):
        """Publishes dashboard or document to the authenticated user's library.

        contents: dashboards or documents to be published, can be
            Dashboard/Document class object or ID
        """
        if not isinstance(contents, list):
            contents = [contents]
        for doc in contents:
            doc_id = doc.id if isinstance(doc, Document) else doc
            body = {'id': doc_id, 'recipients': [{'id': self.user_id}]}
            library.publish_document(self.connection, body=body)

    def unpublish(self, contents: "list | Dashboard | Document | str"):
        """Unpublishes dashboard or document from the authenticated user's
        library.

        contents: dashboards or documents to be published, can be
            Dashboard/Document class object or ID
        """
        if not isinstance(contents, list):
            contents = [contents]
        for doc in contents:
            doc_id = doc.id if isinstance(doc, Document) else doc
            library.unpublish_document_for_user(
                self.connection,
                recipient_id=self.user_id,
                document_id=doc_id,
            )

    def list_library_shortcuts(self) -> list[LibraryShortcut]:
        """List all library shortcuts in the authenticated user's library.

        Returns:
            List of Library items as LibraryShortcut objects.
        """
        self.fetch()
        return [
            LibraryShortcut.from_dict(sho, self.connection)
            for sho in self._doc_dbd_shortcuts_data
            + self._report_shortcuts_data
            + self._agent_shortcuts_data
        ]



"""This is the demo script to show how to manage applications. 
Its basic goal is to present what can be done with this module and to ease
its usage.
"""

from mstrio.connection import get_connection
from mstrio.project_objects.applications import Application, list_applications

conn = get_connection(workstationData, 'MicroStrategy Tutorial')

# list applications with different conditions
# Note: No Applications exist in a default environment
list_of_all_apps = list_applications(connection=conn)
list_of_several_apps = list_applications(connection=conn, limit=5)
list_of_apps_name_filtered = list_applications(connection=conn, name='Strategy')
list_of_all_apps_as_dicts = list_applications(connection=conn, to_dictionary=True)
print(list_of_all_apps)

# Create an Application with customized settings for:
# General Settings, Email Settings, AI Settings, and Auth Modes
# For Home Screen both Library and Document settings are customized
# and the custom Theme settings are set
app = Application.create(
    connection=conn,
    name='Application Validation Scripts',
    description='This is a demo application created by the validation script.',
    home_screen=Application.HomeSettings(
        mode=0,
        home_document=Application.HomeSettings.HomeDocument(
            url='app/B7CA92F04B9FAE8D941C3E9B7E0CD754/DA0B39F911E7A94B00000080AF11FAC4',
            home_document_type='document',
            icons=[
                'table_of_contents',
                'bookmarks',
                'share',
                'options',
            ],
            toolbar_mode=1,
            toolbar_enabled=False,
        ),
        home_library=Application.HomeSettings.HomeLibrary(
            content_bundle_ids=[],
            show_all_contents=None,
            icons=[
                'sidebars',
                'sort_and_filter',
                'search',
                'options',
            ],
            customized_items=Application.HomeSettings.HomeLibrary.CustomizedItems(
                my_content=False,
                subscriptions=True,
                new_dossier=True,
                edit_dossier=True,
                add_library_server=False,
                data_search=True,
                hyper_intelligence=False,
                font_size=True,
                undo_and_redo=True,
                insights=True,
                content_discovery=False,
                mobile_account_panel_user_name=True,
                mobile_account_panel_preferences_my_language=True,
                mobile_account_panel_preferences_my_time_zone=True,
                mobile_account_panel_preferences_face_id_login=False,
                mobile_account_panel_preferences_take_a_tour=True,
                mobile_account_panel_preferences_refresh_view_automatically=True,
                mobile_account_panel_preferences_smart_download=False,
                mobile_account_panel_preferences_automatically_add_to_library=True,
                mobile_account_panel_advanced_settings_app_settings=True,
                mobile_account_panel_advanced_settings_security_settings=True,
                mobile_account_panel_advanced_settings_logging=True,
                mobile_account_panel_help_and_legal=True,
                mobile_account_panel_help_and_legal_help=True,
                mobile_account_panel_help_and_legal_legal=False,
                mobile_account_panel_help_and_legal_report_a_problem=True,
                mobile_account_panel_log_out=True,
                filter_summary=False,
                share_panel_share=True,
                share_panel_export_to_excel=False,
                share_panel_export_to_pdf=True,
                share_panel_download=True,
                share_panel_subscribe=True,
                share_panel_annotate_and_share=True,
                web_account_panel_user_name=True,
                web_account_panel_my_library=False,
                web_account_panel_manage_library=True,
                web_account_panel_preference=True,
                web_account_panel_preference_my_language=True,
                web_account_panel_preference_my_time_zone=True,
                web_account_panel_switch_workspace=False,
                web_account_panel_take_a_tour=True,
                web_account_panel_help=True,
                web_account_panel_log_out=True,
                mobile_downloads=True,
                table_of_contents_header=True,
                table_of_contents_content_info=True,
                table_of_contents_chapter_and_page=True,
                switch_library_server=True,
                create_new_content_dossier=True,
                create_new_content_report=False,
                layout_tile_view=True,
                layout_list_view=True,
                ai_assistant=True,
                share_panel_manage_access=True,
                bot_window_share_panel=True,
                bot_window_share_panel_share_bot=True,
                bot_window_share_panel_embed_bot=True,
                bot_window_share_panel_manage_access=False,
                bot_window_edit_bot=True,
                create_new_content_bot=False,
                dashboard_view_mode=True,
                content_info_content_creator=False,
                content_info_timestamp=True,
                content_info_description=True,
                content_info_project=True,
                content_info_path=True,
                content_info_object_id=True,
                content_info_info_window=True,
                control_filter_summary=True,
                hide_filter_summary=False,
                sidebars_unpin=True,
                table_of_contents_unpin=False,
                filter_panel_unpin=True,
                comments_panel_unpin=False,
                ai_assistant_unpin=True,
                table_of_contents_allow_close=True,
                filter_panel_allow_close=True,
                comments_panel_allow_close=True,
                ai_assistant_allow_close=True,
            ),
            customized_item_properties={
                'report_a_problem': {'address': 'essa@bessa.com'}
            },
            toolbar_mode=1,
            sidebars=[
                'all',
                'favorites',
                'my_groups',
                'options',
            ],
            toolbar_enabled=False,
            default_groups_name='Default Groups Not',
        ),
        theme=Application.HomeSettings.Theme(
            logos=Application.HomeSettings.Theme.Logos(
                web=Application.HomeSettings.Theme.Logos.Logo(
                    type='URL',
                    value=(
                        'https://external-content.duckduckgo.com/iu/?u=https%3A%2F%'
                        'on-D750-2Fnikonrumors.com%2Fwp-content%2Fuploads%2F2014%2F0'
                        '9%2FNiksample-photo1.jpg&f=1&nofb=1&ipt=9639facf02de16ed8'
                        '20d7c3150c2e05820c52e403a54dded4c175d00048dbabe&ipo=images'
                    ),
                ),
                favicon=Application.HomeSettings.Theme.Logos.Logo(
                    type='URL',
                    value=(
                        'https://external-content.duckduckgo.com/iu/?u=https%3A%2F%'
                        'on-D750-2Fnikonrumors.com%2Fwp-content%2Fuploads%2F2014%2F0'
                        '9%2FNiksample-photo1.jpg&f=1&nofb=1&ipt=9639facf02de16ed8'
                        '20d7c3150c2e05820c52e403a54dded4c175d00048dbabe&ipo=images'
                    ),
                ),
                mobile=Application.HomeSettings.Theme.Logos.Logo(
                    type='URL',
                    value=(
                        'https://external-content.duckduckgo.com/iu/?u=https%3A%2F%'
                        'on-D750-2Fnikonrumors.com%2Fwp-content%2Fuploads%2F2014%2F0'
                        '9%2FNiksample-photo1.jpg&f=1&nofb=1&ipt=9639facf02de16ed8'
                        '20d7c3150c2e05820c52e403a54dded4c175d00048dbabe&ipo=images'
                    ),
                ),
            ),
            color=Application.HomeSettings.Theme.Color(
                selected_theme='custom',
                formatting=Application.HomeSettings.Theme.Color.Formatting(
                    toolbar_fill='#0074E7',
                    toolbar_color='#FFFFFF',
                    sidebar_fill='#EDF8FF',
                    sidebar_color='#244D81',
                    sidebar_active_fill='#307BEC',
                    sidebar_active_color='#FFFFFF',
                    panel_fill='#FFFFFF',
                    panel_color='#4E4545',
                    accent_fill='#0077D7',
                    notification_badge_fill='#FF0000',
                    button_color='#FFFFFF',
                    canvas_fill='#F3F4F4',
                ),
                enable_for_bots=False,
            ),
        ),
    ),
    managed=True,
    general=Application.GeneralSettings(
        disable_advanced_settings=False,
        disable_preferences=False,
        network_timeout=420,
        cache_clear_mode=1,
        clear_cache_on_logout=True,
        max_log_size=999,
        log_level=10,
        update_interval=30,
    ),
    platforms=['web', 'mobile', 'desktop'],
    show_builtin_palettes=True,
    is_default=False,
    use_config_palettes=True,
    email_settings=Application.EmailSettings(
        enabled=True,
        host_portal='lanoches.com',
        show_branding_image=True,
        show_browser_button=True,
        show_mobile_button=True,
        show_button_description=True,
        show_reminder=True,
        show_sent_by=True,
        sent_by_text='Strategy One.',
        show_social_media=True,
        content=None,
        sender=Application.EmailSettings.Sender(
            address='library@strategy.com',
            display_name='Strategy One Library',
        ),
        branding_image={
            'url': (
                'https://letsenhance.io/static/8f5e523ee6b2479e26ecc91b9c25'
                '261e/1015f/MainAfter.jpg'
            )
        },
        button=Application.EmailSettings.Button(
            browser_button_style=Application.EmailSettings.Button.ButtonStyle(
                background_color='#3492ed',
                font_color='#ffffff',
                text='View in Browser',
            ),
            mobile_button_style=Application.EmailSettings.Button.ButtonStyle(
                background_color='#3492ed',
                font_color='#ffffff',
                text='View in Mobile App',
            ),
            mobile_button_link_type='DEFAULT',
            mobile_button_scheme='dossier',
            description=(
                '&#39;View in Mobile App&#39; may not work for all mobile '
                'mail apps. Use &#39;View in Browser&#39; option for such '
                'cases.'
            ),
        ),
        reminder=Application.EmailSettings.Reminder(
            text=(
                'Since you last checked, {&amp;NewNotificationCount} notification '
                'has been sent to you.'
            ),
            link_text='View Notification Center',
        ),
        social_media=Application.EmailSettings.SocialMedia(
            show_facebook=True,
            facebook_link='testlink',
            show_twitter=True,
            twitter_link='testlink',
            show_linked_in=True,
            linked_in_link='testlink',
            show_you_tube=True,
            you_tube_link='testlink',
        ),
    ),
    ai_settings=Application.AiSettings(
        feedback=True,
        learning=True,
        disclaimer='Disclaimer 1',
    ),
    auth_modes=Application.AuthModes(
        available_modes=[1, 16],
        default_mode=1,
    ),
)

# Get an Application by its id or name
app = Application(connection=conn, id=app.id)
app = Application(connection=conn, name=app.name)

# Alter an Application
app.alter(
    home_screen=app.home_screen,
    name='Application Validation Scripts Altered',
    general=app.general,
    email_settings=app.email_settings,
    ai_settings=app.ai_settings,
    auth_modes=app.auth_modes,
)

# Delete an Application without a prompt
app.delete(force=True)



from mstrio.api import gateways as gateways_api
from mstrio.connection import Connection
from mstrio.utils.helper import rename_dict_keys

REST_ATTRIBUTES_MAP = {'type': 'gatewayType', 'certifiedAsGateway': 'isCertified'}


def get_all(connection: Connection) -> list[dict]:
    """Get information for all gateways.

    Args:
        connection: Strategy One REST API connection object

    Returns:
        list of dict representing gateway objects
    """
    gateways = gateways_api.get_gateways(connection).json()['gateways'].values()
    return [rename_dict_keys(item, REST_ATTRIBUTES_MAP) for item in gateways]


def get(connection: Connection, id: str) -> dict:
    """Get gateway by a specific ID.

    Args:
        connection: Strategy One REST API connection object
        id: ID of the gateway

    Returns:
        dict representing gateway object
    """
    data = gateways_api.get_gateway(connection, id).json()
    return rename_dict_keys(data, REST_ATTRIBUTES_MAP)



from logging import getLogger

from requests import Response

from mstrio import config
from mstrio.api import timezones as tz_api
from mstrio.connection import Connection
from mstrio.types import ObjectTypes
from mstrio.users_and_groups.user import User
from mstrio.utils import helper
from mstrio.utils.entity import CopyMixin, DeleteMixin, Entity
from mstrio.utils.response_processors import objects as objects_processors
from mstrio.utils.version_helper import class_version_handler, method_version_handler

logger = getLogger(__name__)


@method_version_handler(version='11.3.0500')
def list_timezones(
    connection: Connection,
    to_dictionary: bool = False,
    limit: int | None = None,
    only_system_timezones: bool | None = None,
    only_custom_timezones: bool | None = None,
    **filters,
):
    """Get list of all time zones.

    Args:
        connection: Strategy One connection object returned
            by `connection.Connection()`.
        to_dictionary (bool, optional): If True, returns list of time zones
            as dicts, otherwise list of TimeZone objects. Defaults to False.
        limit (int, optional): The maximum number of time zones to return.
        only_system_timezones (bool, optional): If True, returns only system
            time zones. Cannot be used with `only_custom_timezones`.
        only_custom_timezones (bool, optional): If True, returns only custom
            time zones. Cannot be used with `only_system_timezones`.

    Returns:
        List of time zones as `TimeZone` objects or dictionaries.
    """
    return TimeZone._list_all(
        connection=connection,
        to_dictionary=to_dictionary,
        limit=limit,
        only_system_timezones=only_system_timezones,
        only_custom_timezones=only_custom_timezones,
        **filters,
    )


@class_version_handler(version='11.3.0500')
class TimeZone(Entity, CopyMixin, DeleteMixin):
    _OBJECT_TYPE = ObjectTypes.TIMEZONE
    _API_GETTERS = {
        (
            'name',
            'description',
            'type',
            'subtype',
            'ext_type',
            'date_created',
            'date_modified',
            'version',
            'owner',
            'ancestors',
            'acg',
            'acl',
            'comments',
            'hidden',
        ): objects_processors.get_info,
        (
            'id',
            'base_timezone',
            'current_offset',
        ): tz_api.get_tz,
    }
    _API_PATCH = {
        (
            'name',
            'description',
            'abbreviation',
            'hidden',
            'folder_id',
            'comments',
            'owner',
        ): (objects_processors.update, 'partial_put'),
        ('base_timezone',): (tz_api.update_tz, 'partial_put'),
    }
    _FROM_DICT_MAP = {
        **Entity._FROM_DICT_MAP,
        'base_timezone': lambda source, connection: source['object_id'],
    }

    def __init__(
        self,
        connection: Connection,
        id: str = None,
        name: str = None,
    ):
        """Initialize TimeZone object by passing ID or name.
        When `id` is provided, `name` is omitted.

        Args:
            connection: Strategy One connection object
            id (str, optional): ID of the time zone object
            name (str, optional): Name of the time zone object. Must be provided
                if `id` is not.
        """
        if id is None:
            if name is None:
                raise ValueError(
                    "Please specify either 'name' or 'id' parameter in the constructor."
                )

            tz_obj = helper.find_object_with_name(
                connection=connection,
                cls=self.__class__,
                name=name,
                listing_function=self._list_all,
            )
            object_id = tz_obj['id']
        else:
            object_id = id
        super().__init__(connection=connection, object_id=object_id, name=name)

    def _init_variables(self, **kwargs) -> None:
        super()._init_variables(**kwargs)
        self.name = kwargs.get('name')
        self.current_offset = kwargs.get('current_offset')
        self.base_timezone = kwargs.get('base_timezone')

    @classmethod
    def create(
        cls,
        connection: Connection,
        name: str,
        base_timezone: 'str | TimeZone',
        description: str | None = None,
    ) -> 'TimeZone':
        """Create a new time zone object with the specified properties.

        Args:
            connection: Strategy One connection object returned
                by `connection.Connection()`.
            name (str): Name of the new time zone.
            base_timezone (str | TimeZone): Existing time zone object to base
                the new time zone on.
            description (str, optional): Description of the new time zone.
        """
        if isinstance(base_timezone, cls):
            base_timezone = base_timezone.id
        body = {
            "name": name,
            "description": description,
            "baseTimezone": {
                "objectId": base_timezone,
            },
        }
        body = helper.delete_none_values(source=body, recursion=True)
        res: Response = tz_api.create_tz(connection=connection, body=body)
        new_tz = cls.from_dict(
            source=res.json(),
            connection=connection,
        )
        if config.verbose:
            logger.info(
                f"Successfully created Time Zone named: '{name}' with ID:"
                f" '{new_tz.id}'"
            )
        return new_tz

    def alter(
        self,
        name: str = None,
        base_timezone: 'str | TimeZone | None' = None,
        description: str = None,
        comments: str | None = None,
        owner: str | User | None = None,
    ) -> None:
        """Alter the time zone's properties.
        Args:
            name (str, optional): Name of the time zone object
            base_timezone (str or TimeZone, optional): Reference (ID or TimeZone
                object) to the base time zone
            description (str, optional): Description of the time zone object
            comments (str, optional): long description of the time zone object
            owner: (str or User, optional): owner of the time zone object
        """
        if isinstance(base_timezone, self.__class__):
            base_timezone = base_timezone.id
        if base_timezone is not None:
            base_timezone = {"objectId": base_timezone}
        properties = helper.filter_params_for_func(
            self.alter, locals(), exclude=['self']
        )
        self._alter_properties(**properties)

    @classmethod
    def _list_all(
        cls,
        connection: Connection,
        to_dictionary: bool = False,
        limit: int | None = None,
        only_system_timezones: bool | None = None,
        only_custom_timezones: bool | None = None,
        **filters,
    ) -> list['TimeZone'] | list[dict]:

        if only_system_timezones and only_custom_timezones:
            raise ValueError(
                "You can use only one of "
                "'only_system_timezones' and 'only_custom_timezones' arguments."
            )

        subtype = None
        if only_system_timezones:
            subtype = "timezone_system"
        elif only_custom_timezones:
            subtype = "timezone_custom"

        objects = helper.fetch_objects(
            connection=connection,
            api=tz_api.list_all_tzs,
            limit=limit,
            filters=filters,
            dict_unpack_value='timezones',
            subtype=subtype,
        )
        if to_dictionary:
            return objects
        else:
            return [cls.from_dict(source=obj, connection=connection) for obj in objects]
